"""
Global Raster functions.
These functions are applied to the raster data to create a
processed product on disk, using ImageryLalyer.save() method or arcgis.raster.analytics.generate_raster().

Global functions cannot be used for visualization using dynamic image processing. They cannot be applied to layers that
are added to a map for on-the-fly image processing or visualized inline within the Jupyter notebook.

Functions can be applied to various rasters (or images), including the following:

* Imagery layers
* Rasters within imagery layers

"""
from arcgis.raster._layer import ImageryLayer
from arcgis.gis import Item
import copy
import numbers
from arcgis.raster.functions.utility import _raster_input, _get_raster, _replace_raster_url, _get_raster_url, _get_raster_ra

def _gbl_clone_layer(layer, function_chain, function_chain_ra):
    if isinstance(layer, Item):
        layer = layer.layers[0]    

    newlyr = ImageryLayer(layer._url, layer._gis)

    newlyr._lazy_properties = layer.properties
    newlyr._hydrated = True
    newlyr._lazy_token = layer._token

    # if layer._fn is not None: # chain the functions
    #     old_chain = layer._fn
    #     newlyr._fn = function_chain
    #     newlyr._fn['rasterFunctionArguments']['Raster'] = old_chain
    # else:
    newlyr._fn = function_chain
    newlyr._fnra = function_chain_ra

    newlyr._where_clause = layer._where_clause
    newlyr._spatial_filter = layer._spatial_filter
    newlyr._temporal_filter = layer._temporal_filter
    newlyr._mosaic_rule = layer._mosaic_rule
    newlyr._filtered = layer._filtered
    newlyr._extent = layer._extent
    newlyr._uses_gbl_function = True
    return newlyr


def euclidean_distance(in_source_data,
                       cell_size=None,
                       max_distance=None):
    """
    Calculates, for each cell, the Euclidean distance to the closest source. 
    For more information, see 
    http://pro.arcgis.com/en/pro-app/help/data/imagery/euclidean-distance-global-function.htm

    Parameters
    ----------
    :param in_source_data: raster; The input raster that identifies the pixels or locations to
                            which the Euclidean distance for every output pixel location is calculated.
                            The input type can be an integer or a floating-point value.
    :param cell_size:  The pixel size at which the output raster will be created. If the cell
                            size was explicitly set in Environments, that will be the default cell size. 
                            If Environments was not set, the output cell size will be the same as the 
                            Source Raster
    :param max_distance: The threshold that the accumulative distance values cannot exceed. If an
                            accumulative Euclidean distance exceeds this value, the output value for
                            the pixel location will be NoData. The default distance is to the edge 
                            of the output raster
    :return: output raster with function applied
    """
    layer, in_source_data, raster_ra = _raster_input(in_source_data)
                  
    template_dict = {
        "rasterFunction" : "GPAdapter",
        "rasterFunctionArguments" : {
            "toolName" : "EucDistance_sa",           
            "PrimaryInputParameterName":"in_source_data",
            "OutputRasterParameterName":"out_distance_raster",
            "in_source_data": in_source_data,
                
        }
    }
    
    if cell_size is not None:
        template_dict["rasterFunctionArguments"]["cell_size"] = cell_size
    
    if max_distance is not None:
        template_dict["rasterFunctionArguments"]["maximum_distance"] = max_distance

    function_chain_ra = copy.deepcopy(template_dict)
    function_chain_ra["rasterFunctionArguments"]["in_source_data"] = raster_ra

    return _gbl_clone_layer(layer, template_dict, function_chain_ra)
    


def euclidean_allocation(in_source_data,
                         in_value_raster=None,
                         max_distance=None,
                         cell_size=None,
                         source_field=None):
    
    """
    Calculates, for each cell, the nearest source based on Euclidean distance.
    For more information, see 
    http://pro.arcgis.com/en/pro-app/help/data/imagery/euclidean-allocation-global-function.htm

    Parameters
    ----------
    :param in_source_data: raster; The input raster that identifies the pixels or locations to which
                            the Euclidean distance for every output pixel location is calculated. 
                            The input type can be an integer or a floating-point value.
                            If the input Source Raster is floating point, the Value Raster must be set,
                            and it must be an integer. The Value Raster will take precedence over any
                            setting of the Source Field.
    :param in_value_raster: The input integer raster that identifies the zone values that should be 
                            used for each input source location. For each source location pixel, the
                            value defined by the Value Raster will be assigned to all pixels allocated
                            to the source location for the computation. The Value Raster will take 
                            precedence over any setting for the Source Field .
    :param max_distance: The threshold that the accumulative distance values cannot exceed. If an
                            accumulative Euclidean distance exceeds this value, the output value for
                            the pixel location will be NoData. The default distance is to the edge                     
                            of the output raster
    :param cell_size: The pixel size at which the output raster will be created. If the cell size
                            was explicitly set in Environments, that will be the default cell size. 
                            If Environments was not set, the output cell size will be the same as the 
                            Source Raster
    :param source_field: The field used to assign values to the source locations. It must be an
                            integer type. If the Value Raster has been set, the values in that input
                            will take precedence over any setting for the Source Field.
    :return: output raster with function applied
    """
    
    layer1, in_source_data, raster_ra1 = _raster_input(in_source_data)      
            
    template_dict = {
        "rasterFunction" : "GPAdapter",
        "rasterFunctionArguments" : {
            "toolName" : "EucAllocation_sa",           
            "PrimaryInputParameterName":"in_source_data",
            "OutputRasterParameterName":"out_allocation_raster",
            "in_source_data": in_source_data            
                
        }
    }
    
    if in_value_raster is not None:
        layer2, in_value_raster, raster_ra2 = _raster_input(in_value_raster)
        template_dict["rasterFunctionArguments"]["in_value_raster"] = in_value_raster
    
    if cell_size is not None:
        template_dict["rasterFunctionArguments"]["cell_size"] = cell_size
    
    if max_distance is not None:
        template_dict["rasterFunctionArguments"]["maximum_distance"] = max_distance
    
    if source_field is not None:
        template_dict["rasterFunctionArguments"]["source_field"] = source_field
    
    function_chain_ra = copy.deepcopy(template_dict)
    function_chain_ra['rasterFunctionArguments']["in_source_data"] = raster_ra1
    if in_value_raster is not None:
        function_chain_ra["rasterFunctionArguments"]["in_value_raster"] = raster_ra2

    return _gbl_clone_layer(layer1, template_dict, function_chain_ra)
    
    
    
def cost_distance(in_source_data,
                  in_cost_raster,
                  max_distance=None,
                  source_cost_multiplier=None,
                  source_start_cost=None,
                  source_resistance_rate=None,
                  source_capacity=None,
                  source_direction=None):
    """
    Calculates the least accumulative cost distance for each cell from or to the least-cost
    source over a cost surface.
    For more information, see
    http://pro.arcgis.com/en/pro-app/help/data/imagery/cost-distance-global-function.htm

    Parameters
    ----------
    :param in_source_data: The input raster that identifies the pixels or locations to which the
                            least accumulated cost distance for every output pixel location is 
                            calculated. The Source Raster can be an integer or a floating-point value.
    :param in_cost_raster: A raster defining the cost or impedance to move planimetrically through each pixel.
                            The value at each pixel location represents the cost-per-unit distance for moving 
                            through it. Each pixel location value is multiplied by the pixel resolution, while 
                            also compensating for diagonal movement to obtain the total cost of passing through 
                            the pixel. 
    :param max_distance: The threshold that the accumulative cost values cannot exceed. If an accumulative cost
                            distance exceeds this value, the output value for the pixel location will be NoData. 
                            The maximum distance defines the extent for which the accumulative cost distances are
                            calculated. The default distance is to the edge of the output raster.
    :param source_cost_multiplier: The threshold that the accumulative cost values cannot exceed. If an accumulative
                            cost distance exceeds this value, the output value for the pixel location will be 
                            NoData. The maximum distance defines the extent for which the accumulative cost 
                            distances are calculated. The default distance is to the edge of the output raster.
    :param source_start_cost: The starting cost from which to begin the cost calculations. This parameter allows
                            for the specification of the fixed cost associated with a source. Instead of starting
                            at a cost of 0, the cost algorithm will begin with the value set here.
                            The default is 0. The value must be 0 or greater. A numeric (double) value or a field
                            from the Source Raster can be used for this parameter.
    :param source_resistance_rate: This parameter simulates the increase in the effort to overcome costs as the
                            accumulative cost increases. It is used to model fatigue of the traveler. The growing
                            accumulative cost to reach a pixel is multiplied by the resistance rate and added to 
                            the cost to move into the subsequent pixel.
                            It is a modified version of a compound interest rate formula that is used to calculate
                            the apparent cost of moving through a pixel. As the value of the resistance rate increases,
                            it increases the cost of the pixels that are visited later. The greater the resistance rate, 
                            the higher the cost to reach the next pixel, which is compounded for each subsequent movement. 
                            Since the resistance rate is similar to a compound rate and generally the accumulative cost 
                            values are very large, small resistance rates are suggested, such as 0.005 or even smaller, 
                            depending on the accumulative cost values.
                            The default is 0. The values must be 0 or greater. A numeric (double) value or a field from
                            the Source Raster can be used for this parameter.
    :param source_capacity: Defines the cost capacity for the traveler for a source. The cost calculations continue for
                            each source until the specified capacity is reached.
                            The default capacity is to the edge of the output raster. The values must be greater than 0. 
                            A double numeric value or a field from the Source Raster can be used for this parameter.
    :param source_direction: Defines the direction of the traveler when applying the source resistance rate and the source
                            starting cost.
                            From Source - The source resistance rate and source starting cost will be applied beginning
                            at the input source and moving out to the nonsource cells. This is the default.
                            To Source - The source resistance rate and source starting cost will be applied beginning at
                            each nonsource cell and moving back to the input source.
                            Either specify the From Source or To Source keyword, which will be applied to all sources,
                            or specify a field in the Source Raster that contains the keywords to identify the direction
                            of travel for each source. That field must contain the string From Source or To Source.
    
    :return: output raster with function applied
    """        
    layer1, in_source_data, raster_ra1 = _raster_input(in_source_data)  
    layer2, in_cost_raster, raster_ra2 = _raster_input(in_cost_raster)
                            
    template_dict = {
        "rasterFunction" : "GPAdapter",
        "rasterFunctionArguments" : {
            "toolName" : "CostDistance_sa",           
            "PrimaryInputParameterName":"in_source_data",
            "OutputRasterParameterName":"out_distance_raster",
            "in_source_data": in_source_data, 
            "in_cost_raster": in_cost_raster
             
        }
    }    
    
    if max_distance is not None:
        template_dict["rasterFunctionArguments"]["maximum_distance"] = max_distance
    
    if source_cost_multiplier is not None:
        template_dict["rasterFunctionArguments"]["source_cost_multiplier"] = source_cost_multiplier
    
    if source_start_cost is not None:
        template_dict["rasterFunctionArguments"]["source_start_cost"] = source_start_cost
    
    if source_resistance_rate is not None:
        template_dict["rasterFunctionArguments"]["source_resistance_rate"] = source_resistance_rate
    
    if source_capacity is not None:
        template_dict["rasterFunctionArguments"]["source_capacity"] = source_capacity
    
    if source_direction is not None:
        template_dict["rasterFunctionArguments"]["source_direction"] = source_direction
    
    function_chain_ra = copy.deepcopy(template_dict)
    function_chain_ra["rasterFunctionArguments"]["in_source_data"] = raster_ra1
    function_chain_ra["rasterFunctionArguments"]["in_cost_raster"] = raster_ra2

    return _gbl_clone_layer(layer1, template_dict, function_chain_ra)
        
        
def cost_allocation(in_source_data,
                    in_cost_raster,
                    in_value_raster=None,
                    max_distance=None,                    
                    source_field=None,
                    source_cost_multiplier=None,
                    source_start_cost=None,
                    source_resistance_rate=None,
                    source_capacity=None,
                    source_direction=None):
    """
    Calculates, for each cell, its least-cost source based on the least accumulative cost over a cost surface.
    For more information, see
    http://pro.arcgis.com/en/pro-app/help/data/imagery/cost-allocation-global-function.htm

    Parameters
    ----------
    :param in_source_data: The input raster that identifies the pixels or locations to which the
                            least accumulated cost distance for every output pixel location is 
                            calculated. The Source Raster can be an integer or a floating-point value.
                            If the input Source Raster is floating point, the Value Raster must be set, 
                            and it must be an integer. The Value Raster will take precedence over any 
                            setting of the Source Field.
    :param in_cost_raster: A raster defining the cost or impedance to move planimetrically through each pixel.
                            The value at each pixel location represents the cost-per-unit distance for moving 
                            through it. Each pixel location value is multiplied by the pixel resolution, while 
                            also compensating for diagonal movement to obtain the total cost of passing through 
                            the pixel. 
                            The values of the Cost Raster can be integer or floating point, but they cannot be 
                            negative or zero.
    :param in_value_raster: The input integer raster that identifies the zone values that should be used for 
                            each input source location. For each source location pixel, the value defined by
                            the Value Raster will be assigned to all pixels allocated to the source location 
                            for the computation. The Value Raster will take precedence over any setting for 
                            the Source Field. 
    :param max_distance: The threshold that the accumulative cost values cannot exceed. If an accumulative cost
                            distance exceeds this value, the output value for the pixel location will be NoData. 
                            The maximum distance defines the extent for which the accumulative cost distances are
                            calculated. The default distance is to the edge of the output raster.
    :param source_field: The field used to assign values to the source locations. It must be an integer type.
                            If the Value Raster has been set, the values in that input will take precedence over
                            any setting for the Source Field.
    :param source_cost_multiplier: This parameter allows for control of the mode of travel or the magnitude at
                            a source. The greater the multiplier, the greater the cost to move through each cell.
                            The default value is 1. The values must be greater than 0. A numeric (double) value or
                            a field from the Source Raster can be used for this parameter.
    :param source_start_cost: The starting cost from which to begin the cost calculations. This parameter allows
                            for the specification of the fixed cost associated with a source. Instead of starting
                            at a cost of 0, the cost algorithm will begin with the value set here.
                            The default is 0. The value must be 0 or greater. A numeric  (double) value or a field
                            from the Source Raster can be used for this parameter.
    :param source_resistance_rate: This parameter simulates the increase in the effort to overcome costs as the
                            accumulative cost increases. It is used to model fatigue of the traveler. The growing
                            accumulative cost to reach a pixel is multiplied by the resistance rate and added to 
                            the cost to move into the subsequent pixel.
                            It is a modified version of a compound interest rate formula that is used to calculate
                            the apparent cost of moving through a pixel. As the value of the resistance rate increases,
                            it increases the cost of the pixels that are visited later. The greater the resistance rate, 
                            the higher the cost to reach the next pixel, which is compounded for each subsequent movement. 
                            Since the resistance rate is similar to a compound rate and generally the accumulative cost 
                            values are very large, small resistance rates are suggested, such as 0.005 or even smaller, 
                            depending on the accumulative cost values.
                            The default is 0. The values must be 0 or greater. A numeric (double) value or a field from
                            the Source Raster can be used for this parameter.
    :param source_capacity: Defines the cost capacity for the traveler for a source. The cost calculations continue for
                            each source until the specified capacity is reached.
                            The default capacity is to the edge of the output raster. The values must be greater than 0. 
                            A double numeric value or a field from the Source Raster can be used for this parameter.
    :source_direction: Defines the direction of the traveler when applying the source resistance rate and the source
                            starting cost.
                            From Source - The source resistance rate and source starting cost will be applied beginning
                            at the input source and moving out to the nonsource cells. This is the default.
                            To Source - The source resistance rate and source starting cost will be applied beginning at
                            each nonsource cell and moving back to the input source.
                            Either specify the From Source or To Source keyword, which will be applied to all sources,
                            or specify a field in the Source Raster that contains the keywords to identify the direction
                            of travel for each source. That field must contain the string From Source or To Source.
    
    :return: output raster with function applied
    """   
            
    layer1, in_source_data, raster_ra1 = _raster_input(in_source_data)
    layer2, in_cost_raster, raster_ra2 = _raster_input(in_cost_raster)
            
    template_dict = {
        "rasterFunction" : "GPAdapter",
        "rasterFunctionArguments" : {
            "toolName" : "CostAllocation_sa",           
            "PrimaryInputParameterName":"in_source_data",
            "OutputRasterParameterName":"out_allocation_raster",
            "in_source_data": in_source_data, 
            "in_cost_raster": in_cost_raster
             
        }
    }    
    if in_value_raster is not None:
        layer3, in_value_raster, raster_ra3 = _raster_input(in_value_raster)
        template_dict["rasterFunctionArguments"]["in_value_raster"] = in_value_raster

    if max_distance is not None:
        template_dict["rasterFunctionArguments"]["maximum_distance"] = max_distance

    if source_field is not None:
        template_dict["rasterFunctionArguments"]["source_field"] = source_field
    
    if source_cost_multiplier is not None:
        template_dict["rasterFunctionArguments"]["source_cost_multiplier"] = source_cost_multiplier
    
    if source_start_cost is not None:
        template_dict["rasterFunctionArguments"]["source_start_cost"] = source_start_cost
    
    if source_resistance_rate is not None:
        template_dict["rasterFunctionArguments"]["source_resistance_rate"] = source_resistance_rate
    
    if source_capacity is not None:
        template_dict["rasterFunctionArguments"]["source_capacity"] = source_capacity
    
    if source_direction is not None:
        template_dict["rasterFunctionArguments"]["source_direction"] = source_direction
    
    function_chain_ra = copy.deepcopy(template_dict)
    function_chain_ra["rasterFunctionArguments"]["in_source_data"] = raster_ra1
    function_chain_ra["rasterFunctionArguments"]["in_cost_raster"] = raster_ra2
    if in_value_raster is not None:
        function_chain_ra["rasterFunctionArguments"]["in_value_raster"] = raster_ra3

    return _gbl_clone_layer(layer1, template_dict, function_chain_ra)
    
    
def zonal_statistics(in_zone_data,
                     zone_field,
                     in_value_raster,
                     ignore_nodata=None,
                     statistics_type=None):
                     
    """"
    Calculates statistics on values of a raster within the zones of another dataset.
    For more information, 
    http://pro.arcgis.com/en/pro-app/help/data/imagery/zonal-statistics-global-function.htm

    Parameters
    ----------
    :param in_zone_data: Dataset that defines the zones. The zones can be defined by an integer raster
    :param zone_field: Field that holds the values that define each zone. It can be an integer or a
                            string field of the zone raster.
    :param in_value_raster: Raster that contains the values on which to calculate a statistic.
    :param ignore_no_data: Denotes whether NoData values in the Value Raster will influence the results
                            of the zone that they fall within.
                            Yes - Within any particular zone, only pixels that have a value in the Value
                            Raster will be used in determining the output value for that zone. NoData 
                            pixels in the Value Raster will be ignored in the statistic calculation. 
                            This is the default.
                            No - Within any particular zone, if any NoData pixels exist in the Value 
                            Raster, it is deemed that there is insufficient information to perform 
                            statistical calculations for all the pixels in that zone; therefore, the 
                            entire zone will receive the NoData value on the output raster.
    :param statistics_type: Statistic type to be calculated.
                            Mean-Calculates the average of all pixels in the Value Raster that belong to
                            the same zone as the output pixel.
                            Majority-Determines the value that occurs most often of all pixels in the 
                            Value Raster that belong to the same zone as the output pixel.
                            Maximum-Determines the largest value of all pixels in the Value Raster 
                            that belong to the same zone as the output pixel.
                            Median-Determines the median value of all pixels in the Value Raster
                            that belong to the same zone as the output pixel.
                            Minimum-Determines the smallest value of all pixels in the Value Raster 
                            that belong to the same zone as the output pixel.
                            Minority-Determines the value that occurs least often of all pixels in
                            the Value Raster that belong to the same zone as the output pixel.
                            Range-Calculates the difference between the largest and smallest value 
                            of all pixels in the Value Raster that belong to the same zone as the
                            output pixel.
                            Standard Deviation-Calculates the standard deviation of all pixels in
                            the Value Rasterthat belong to the same zone as the output pixel.
                            Sum-Calculates the total value of all pixels in the Value Raster that
                            belong to the same zone as the output pixel.
                            Variety-Calculates the number of unique values for all pixels in the 
                            Value Raster that belong to the same zone as the output pixel.
    :return: output raster with function applied
    """
    layer1, in_zone_data, raster_ra1 = _raster_input(in_zone_data)  
    layer2, in_value_raster, raster_ra2 = _raster_input(in_value_raster)
        
    template_dict = {
        "rasterFunction" : "GPAdapter",
        "rasterFunctionArguments" : {
            "toolName" : "ZonalStatistics_sa",           
            "PrimaryInputParameterName" : "in_zone_data",
            "OutputRasterParameterName" : "out_raster",
            "in_zone_data" : in_zone_data, 
            "zone_field" : zone_field,
            "in_value_raster" : in_value_raster
             
        }
    }    
    
    if ignore_nodata is not None:
        template_dict["rasterFunctionArguments"]["ignore_nodata"] = ignore_nodata
    
    if statistics_type is not None:
        template_dict["rasterFunctionArguments"]["statistics_type"] = statistics_type
    
    function_chain_ra = copy.deepcopy(template_dict)
    function_chain_ra['rasterFunctionArguments']["in_zone_data"] = raster_ra1
    function_chain_ra["rasterFunctionArguments"]["in_value_raster"] = raster_ra2

    return _gbl_clone_layer(layer1, template_dict, function_chain_ra)        
        
        
def least_cost_path(in_source_data,
                    in_cost_raster,
                    in_destination_data,
                    destination_field=None,                    
                    path_type=None,
                    max_distance=None,
                    source_cost_multiplier=None,
                    source_start_cost=None,
                    source_resistance_rate=None,
                    source_capacity=None,
                    source_direction=None):
    """
    Calculates the least-cost path from a source to a destination. The least accumulative cost distance 
    is calculated for each pixel over a cost surface, to the nearest source. This produces an output 
    raster that records the least-cost path, or paths, from selected locations to the closest source
    pixels defined within the accumulative cost surface, in terms of cost distance.
    For more information, see
    http://pro.arcgis.com/en/pro-app/help/data/imagery/least-cost-path-global-function.htm

    Parameters
    ----------
    :param in_source_data: The input raster that identifies the pixels or locations to which the
                            least accumulated cost distance for every output pixel location is 
                            calculated. The Source Raster can be an integer or a floating-point value.
                            If the input Source Raster is floating point, the Value Raster must be set, 
                            and it must be an integer. The Value Raster will take precedence over any 
                            setting of the Source Field.
    :param in_cost_raster: A raster defining the cost or impedance to move planimetrically through each pixel.
                            The value at each pixel location represents the cost-per-unit distance for moving 
                            through it. Each pixel location value is multiplied by the pixel resolution, while 
                            also compensating for diagonal movement to obtain the total cost of passing through 
                            the pixel. 
                            The values of the Cost Raster can be integer or floating point, but they cannot be 
                            negative or zero.
    :param in_destination_data: A raster dataset that identifies the pixels from which the least-cost path is 
                            determined to the least costly source. This input consists of pixels that have valid
                            values, and the remaining pixels must be assigned NoData. Values of 0 are valid.
    :param destination_field: The field used to obtain values for the destination locations.
    :param path_type: A keyword defining the manner in which the values and zones on the input destination
                            data will be interpreted in the cost path calculations:
                            EACH_CELL-A least-cost path is determined for each pixel with valid values on the 
                            input destination data, and saved on the output raster. Each cell of the input 
                            destination data is treated separately, and a least-cost path is determined for each from cell.
                            EACH_ZONE-A least-cost path is determined for each zone on the input destination data and
                            saved on the output raster. The least-cost path for each zone begins at the pixel with the
                            lowest cost distance weighting in the zone.
                            BEST_SINGLE-For all pixels on the input destination data, the least-cost path is derived
                            from the pixel with the minimum of the least-cost paths to source cells.
    :param max_distance: The threshold that the accumulative cost values cannot exceed. If an accumulative cost
                            distance exceeds this value, the output value for the pixel location will be NoData. 
                            The maximum distance defines the extent for which the accumulative cost distances are
                            calculated. The default distance is to the edge of the output raster.
    :param source_field: The field used to assign values to the source locations. It must be an integer type.
                            If the Value Raster has been set, the values in that input will take precedence over
                            any setting for the Source Field.
    :param source_cost_multiplier: The threshold that the accumulative cost values cannot exceed. If an accumulative
                            cost distance exceeds this value, the output value for the pixel location will be 
                            NoData. The maximum distance defines the extent for which the accumulative cost 
                            distances are calculated. The default distance is to the edge of the output raster.
    :param source_start_cost: The starting cost from which to begin the cost calculations. This parameter allows
                            for the specification of the fixed cost associated with a source. Instead of starting
                            at a cost of 0, the cost algorithm will begin with the value set here.
                            The default is 0. The value must be 0 or greater. A numeric (double) value or a field
                            from the Source Raster can be used for this parameter.
    :param source_resistance_rate: This parameter simulates the increase in the effort to overcome costs as the
                            accumulative cost increases. It is used to model fatigue of the traveler. The growing
                            accumulative cost to reach a pixel is multiplied by the resistance rate and added to 
                            the cost to move into the subsequent pixel.
                            It is a modified version of a compound interest rate formula that is used to calculate
                            the apparent cost of moving through a pixel. As the value of the resistance rate increases,
                            it increases the cost of the pixels that are visited later. The greater the resistance rate, 
                            the higher the cost to reach the next pixel, which is compounded for each subsequent movement. 
                            Since the resistance rate is similar to a compound rate and generally the accumulative cost 
                            values are very large, small resistance rates are suggested, such as 0.005 or even smaller, 
                            depending on the accumulative cost values.
                            The default is 0. The values must be 0 or greater. A numeric (double) value or a field from
                            the Source Raster can be used for this parameter.
    :param source_capacity: Defines the cost capacity for the traveler for a source. The cost calculations continue for
                            each source until the specified capacity is reached.
                            The default capacity is to the edge of the output raster. The values must be greater than 0. 
                            A double numeric value or a field from the Source Raster can be used for this parameter.
    :param source_direction: Defines the direction of the traveler when applying the source resistance rate and the source
                            starting cost.
                            From Source - The source resistance rate and source starting cost will be applied beginning
                            at the input source and moving out to the nonsource cells. This is the default.
                            To Source-The source resistance rate and source starting cost will be applied beginning at
                            each nonsource cell and moving back to the input source.
                            Either specify the From Source or To Source keyword, which will be applied to all sources,
                            or specify a field in the Source Raster that contains the keywords to identify the direction
                            of travel for each source. That field must contain the string From Source or To Source.
    
    :return: output raster with function applied
    """  
    layer1, in_source_data, raster_ra1 = _raster_input(in_source_data)
    layer2, in_cost_raster, raster_ra2 = _raster_input(in_cost_raster)
    layer3, in_destination_data, raster_ra3 = _raster_input(in_destination_data)
            
    template_dict = {
        "rasterFunction" : "GPAdapter",
        "rasterFunctionArguments" : {
            "toolName" : "ShortestPath",           
            "PrimaryInputParameterName" : "in_source_data",
            "OutputRasterParameterName":"out_path_raster",
            "in_source_data" : in_source_data, 
            "in_cost_raster" : in_cost_raster,
            "in_destination_data" : in_destination_data
                
        }
    }    
            
    if destination_field is not None:
        template_dict["rasterFunctionArguments"]["destination_field"] = destination_field
    
    if path_type is not None:
        template_dict["rasterFunctionArguments"]["path_type"] = path_type
    
    if max_distance is not None:
        template_dict["rasterFunctionArguments"]["maximum_distance"] = max_distance
    
    if source_cost_multiplier is not None:
        template_dict["rasterFunctionArguments"]["source_cost_multiplier"] = source_cost_multiplier
    
    if source_start_cost is not None:
        template_dict["rasterFunctionArguments"]["source_start_cost"] = source_start_cost
    
    if source_resistance_rate is not None:
        template_dict["rasterFunctionArguments"]["source_resistance_rate"] = source_resistance_rate
    
    if source_capacity is not None:
        template_dict["rasterFunctionArguments"]["source_capacity"] = source_capacity
    
    if source_direction is not None:
        template_dict["rasterFunctionArguments"]["source_direction"] = source_direction

    function_chain_ra = copy.deepcopy(template_dict)
    function_chain_ra['rasterFunctionArguments']["in_source_data"] = raster_ra1
    function_chain_ra["rasterFunctionArguments"]["in_cost_raster"] = raster_ra2
    function_chain_ra["rasterFunctionArguments"]["in_destination_data"] = raster_ra3

    return _gbl_clone_layer(layer1, template_dict, function_chain_ra)


def flow_distance(input_stream_raster,
                  input_surface_raster,
                  input_flow_direction_raster=None,
                  distance_type=None):
                     
    """
    This function computes, for each cell, the minimum downslope 
    horizontal or vertical distance to cell(s) on a stream or
    river into which they flow. If an optional flow direction 
    raster is provided, the down slope direction(s) will be 
    limited to those defined by the input flow direction raster.

    Parameters
    ----------
    :param input_stream_raster: An input raster that represents a linear stream network
    :param input_surface_raster: The input raster representing a continuous surface.
    :param input_flow_direction_raster: The input raster that shows the direction of flow out of each cell.
    :param distance_type: VERTICAL or HORIZONTAL distance to compute; if not
                                 specified, VERTICAL distance is computed.
    :return: output raster with function applied
    """
    layer1, input_stream_raster, raster_ra1 = _raster_input(input_stream_raster)  
    layer2, input_surface_raster, raster_ra2 = _raster_input(input_surface_raster)

        
    template_dict = {
        "rasterFunction" : "GPAdapter",
        "rasterFunctionArguments" : {
            "toolName" : "FlowDistance_sa",           
            "PrimaryInputParameterName" : "in_stream_raster",
            "OutputRasterParameterName" : "out_raster",
            "in_stream_raster" : input_stream_raster, 
            "in_surface_raster" : input_surface_raster,
             
        }
    }    
    if input_flow_direction_raster is not None:
        layer3, input_flow_direction_raster, raster_ra3 = _raster_input(input_flow_direction_raster)
        template_dict["rasterFunctionArguments"]["in_flow_direction_raster"] = input_flow_direction_raster

    if distance_type is not None:
        template_dict["rasterFunctionArguments"]["distance_type"] = distance_type
    
    function_chain_ra = copy.deepcopy(template_dict)
    function_chain_ra['rasterFunctionArguments']["in_stream_raster"] = raster_ra1
    function_chain_ra['rasterFunctionArguments']["in_surface_raster"] = raster_ra2
    if input_flow_direction_raster is not None:
        function_chain_ra["rasterFunctionArguments"]["in_flow_direction_raster"] = raster_ra3

    return _gbl_clone_layer(layer1, template_dict, function_chain_ra)


def flow_accumulation(input_flow_direction_raster,
                      input_weight_raster=None,
                      data_type=None):
                     
    """"    
    Replaces cells of a raster corresponding to a mask 
    with the values of the nearest neighbors.

    :param input_flow_direction_raster: The input raster that shows the direction of flow out of each cell.
    :param input_weight_raster: An optional input raster for applying a weight to each cell.
    :param data_type: INTEGER, FLOAT, DOUBLE
    :return: output raster with function applied

    """
    layer1, input_flow_direction_raster, raster_ra1 = _raster_input(input_flow_direction_raster)  
            
    template_dict = {
        "rasterFunction" : "GPAdapter",
        "rasterFunctionArguments" : {
            "toolName" : "FlowAccumulation_sa",           
            "PrimaryInputParameterName" : "in_flow_direction_raster",
            "OutputRasterParameterName" : "out_accumulation_raster",
            "in_flow_direction_raster" : input_flow_direction_raster
             
        }
    }    
    if input_weight_raster is not None:
        layer2, input_weight_raster, raster_ra2 = _raster_input(input_weight_raster) 
        template_dict["rasterFunctionArguments"]["in_weight_raster"] = input_weight_raster

    if data_type is not None:
        template_dict["rasterFunctionArguments"]["data_type"] = data_type
    
    function_chain_ra = copy.deepcopy(template_dict)
    function_chain_ra["rasterFunctionArguments"]["in_flow_direction_raster"] = raster_ra1    
    if input_weight_raster is not None:
        function_chain_ra['rasterFunctionArguments']["in_weight_raster"] = raster_ra2    

    return _gbl_clone_layer(layer1, template_dict, function_chain_ra)


def flow_direction(input_surface_raster,
                   force_flow= "NORMAL",
                   flow_direction_type= "D8"):
    """
    Replaces cells of a raster corresponding to a mask 
    with the values of the nearest neighbors.

    Parameters
    ----------
    :param input_surface_raster: The input raster representing a continuous surface.
    :param force_flow: NORMAL or FORCE, Specifies if edge cells will always flow outward or follow normal flow rules.
    :param flow_direction_type: Specifies which flow direction type to use.
                          D8 - Use the D8 method. This is the default.
                          MFD - Use the Multi Flow Direction (MFD) method.
                          DINF - Use the D-Infinity method.
    
    :return: output raster with function applied

    """
    layer, input_surface_raster, raster_ra = _raster_input(input_surface_raster)  
            
    template_dict = {
        "rasterFunction" : "GPAdapter",
        "rasterFunctionArguments" : {
            "toolName" : "FlowDirection_sa",           
            "PrimaryInputParameterName" : "in_surface_raster",
            "OutputRasterParameterName" : "out_flow_direction_raster",
            "in_surface_raster" : input_surface_raster,
            "force_flow" : force_flow,
            "flow_direction_type" : flow_direction_type
             
        }
    }    
        
    function_chain_ra = copy.deepcopy(template_dict)
    function_chain_ra["rasterFunctionArguments"]["in_surface_raster"] = raster_ra    
    
    return _gbl_clone_layer(layer, template_dict, function_chain_ra)


def fill(input_surface_raster,        
         zlimit=None):
    """
    Fills sinks in a surface raster to remove small imperfections in the data

    Parameters
    ----------
    :param input_surface_raster: The input raster representing a continuous surface.
    :param zlimit: Data type - Double. Maximum elevation difference between a sink and
            its pour point to be filled.
            If the difference in z-values between a sink and its pour point is greater than the z_limit, that sink will not be filled.
            The value for z-limit must be greater than zero.
            Unless a value is specified for this parameter, all sinks will be filled, regardless of depth.
    
    :return: output raster with function applied

    """
    layer, input_surface_raster, raster_ra = _raster_input(input_surface_raster)  
            
    template_dict = {
        "rasterFunction" : "GPAdapter",
        "rasterFunctionArguments" : {
            "toolName" : "Fill_sa",           
            "PrimaryInputParameterName" : "in_surface_raster",
            "OutputRasterParameterName" : "out_surface_raster",
            "in_surface_raster" : input_surface_raster            
             
        }
    }    
    
    if zlimit is not None:
        template_dict["rasterFunctionArguments"]["z_limit"] = zlimit
    
    function_chain_ra = copy.deepcopy(template_dict)
    function_chain_ra["rasterFunctionArguments"]["in_surface_raster"] = raster_ra    
    
    return _gbl_clone_layer(layer, template_dict, function_chain_ra)


def nibble(input_raster,
           input_mask_raster,
           nibble_values= "ALL_VALUES",
           nibble_no_data= "PRESERVE_NODATA",
           input_zone_raster=None):
    """
    Replaces cells of a raster corresponding to a mask 
    with the values of the nearest neighbors.

    Parameters
    ----------
    :param input_raster: The input rater to nibble.
                   The input raster can be either integer or floating point type.
    :param input_mask_raster: The input raster to use as the mask.
    :param nibble_values: possbile options are "ALL_VALUES" and "DATA_ONLY".
        Default is "ALL_VALUES"
    :param nibble_no_data: PRESERVE_NODATA or PROCESS_NODATA possible values;
        Default is PRESERVE_NODATA.
    :param input_zone_raster: The input raster that defines the zones to use as the mask.
    
    :return: output raster with function applied

    """
    layer1, input_raster, raster_ra1 = _raster_input(input_raster)
    layer2, input_mask_raster, raster_ra2 = _raster_input(input_mask_raster)  
            
    template_dict = {
        "rasterFunction" : "GPAdapter",
        "rasterFunctionArguments" : {
            "toolName" : "Nibble_sa",           
            "PrimaryInputParameterName" : "in_raster",
            "OutputRasterParameterName" : "out_raster",
            "in_raster" : input_raster,
            "in_mask_raster" : input_mask_raster,
            "nibble_values" : nibble_values,
            "nibble_nodata" : nibble_no_data
             
        }
    }    
    
    if input_zone_raster is not None:
        layer3, input_zone_raster, raster_ra3 = _raster_input(input_zone_raster)
        template_dict["rasterFunctionArguments"]["in_zone_raster"] = input_zone_raster
    
    function_chain_ra = copy.deepcopy(template_dict)
    function_chain_ra["rasterFunctionArguments"]["in_raster"] = raster_ra1
    function_chain_ra["rasterFunctionArguments"]["in_mask_raster"] = raster_ra2
    if input_zone_raster is not None:
        function_chain_ra["rasterFunctionArguments"]["in_zone_raster"] = raster_ra3
    
    return _gbl_clone_layer(layer1, template_dict, function_chain_ra)


def stream_link(input_raster,
                input_flow_direction_raster):
    """
    Assigns unique values to sections of a raster linear network between intersections

    Parameters
    ----------
    :param input_raster:     An input raster that represents a linear stream network.
    :param input_flow_direction_raster: The input raster that shows the direction of flow out of each cell
    :return: output raster with function applied

    """
    layer1, input_raster, raster_ra1 = _raster_input(input_raster)
    layer2, input_flow_direction_raster, raster_ra2 = _raster_input(input_flow_direction_raster)  
            
    template_dict = {
        "rasterFunction" : "GPAdapter",
        "rasterFunctionArguments" : {
            "toolName" : "StreamLink_sa",           
            "PrimaryInputParameterName" : "in_stream_raster",
            "OutputRasterParameterName" : "out_raster",
            "in_stream_raster" : input_raster,
            "in_flow_direction_raster" : input_flow_direction_raster                 
        }
    }    
    
    
    function_chain_ra = copy.deepcopy(template_dict)
    function_chain_ra["rasterFunctionArguments"]["in_stream_raster"] = raster_ra1
    function_chain_ra["rasterFunctionArguments"]["in_flow_direction_raster"] = raster_ra2
        
    return _gbl_clone_layer(layer1, template_dict, function_chain_ra)


def watershed(input_flow_direction_raster,
              input_pour_point_data,
              pour_point_field=None):
    """
    Replaces cells of a raster corresponding to a mask 
    with the values of the nearest neighbors.

    Parameters
    ----------
    :param input_flow_direction_raster: The input raster that shows the direction of flow out of each cell.
    :param input_pour_point_data: The input pour point locations. For a raster, this represents cells above
                            which the contributing area, or catchment, will be determined. All cells that 
                            are not NoData will be used as source cells.
                            For a point feature dataset, this represents locations above which the contributing
                            area, or catchment, will be determined.
    :param pour_point_field: Field used to assign values to the pour point locations. If the pour point dataset is a
                       raster, use Value.
                       If the pour point dataset is a feature, use a numeric field. If the field contains 
                       floating-point values, they will be truncated into integers.    
    :return: output raster with function applied

    """
    layer1, input_flow_direction_raster, raster_ra1 = _raster_input(input_flow_direction_raster)  
    layer2, input_pour_point_data, raster_ra2 = _raster_input(input_pour_point_data)
            
    template_dict = {
        "rasterFunction" : "GPAdapter",
        "rasterFunctionArguments" : {
            "toolName" : "Watershed_sa",           
            "PrimaryInputParameterName" : "in_flow_direction_raster",
            "OutputRasterParameterName" : "out_raster",
            "in_flow_direction_raster" : input_flow_direction_raster,
            "in_pour_point_data" : input_pour_point_data
        }
    }    
    
    if pour_point_field is not None:
        template_dict["rasterFunctionArguments"]["pour_point_field"] = pour_point_field

    
    function_chain_ra = copy.deepcopy(template_dict)
    function_chain_ra["rasterFunctionArguments"]["in_flow_direction_raster"] = raster_ra1
    function_chain_ra["rasterFunctionArguments"]["in_pour_point_data"] = raster_ra2
        
    return _gbl_clone_layer(layer1, template_dict, function_chain_ra)



"""
Spatial DataFrame Object developed off of the Panda's Dataframe object
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import warnings

import arcgis
from six import string_types, integer_types


HAS_PANDAS = True
try:
    import pandas as pd
    from pandas import DataFrame, Series, Index
    import numpy
    from .base import BaseSpatialPandas
    from .geoseries import GeoSeries
except:
    HAS_PANDAS = False

    class DataFrame:
        pass

    class BaseSpatialPandas:
        pass

from arcgis.gis import GIS
from six import PY3
from six import string_types
from arcgis.geometry import _types
GEO_COLUMN_DEFAULT = "SHAPE"
GEOM_TYPES = (_types.Point, _types.MultiPoint,
              _types.Polygon,_types.Geometry,
              _types.Polyline,
              _types.BaseGeometry)

try:
    import arcpy
    from arcpy import Geometry
    HASARCPY = True
    HAS_ARCPY = True
    GEOM_TYPES = [arcpy.Point, arcpy.Polygon,
                  arcpy.Geometry, arcpy.PointGeometry,
                  arcpy.Polyline, arcpy.Multipatch,
                  arcpy.Multipoint] + list(GEOM_TYPES)
    GEOM_TYPES = tuple(GEOM_TYPES)
except ImportError:
    # warning.warn("Missing Pro will cause functionality to be limited")
    HASARCPY = False
    HAS_ARCPY = False
try:
    import shapely
    from shapely.geometry.base import BaseGeometry as _BaseGeometry
    GEOM_TYPES = [_BaseGeometry] + list(GEOM_TYPES)
    GEOM_TYPES = tuple(GEOM_TYPES)
    HASSHAPELY = True
except:
    HASSHAPELY = False

class SpatialDataFrame(BaseSpatialPandas, DataFrame):
    """
        A Spatial Dataframe is an object to manipulate, manage and translate
        data into new forms of information for users.

        Functionality of the Spatial DataFrame is determined by the Geometry Engine
        available to the object at creation.  It will first leverage the arcpy
        geometry engine, then shapely, then it will create the geometry objects
        without any engine.

        **Scenerios**

        =================   ======================================================
        **Engine Type**     **Functionality**
        -----------------   ------------------------------------------------------
        ArcPy               Users will have the full functionality provided by the
                            API.
        -----------------   ------------------------------------------------------
        Shapely             Users get a sub-set of operations, and all properties.

                            :Valid Properties:

                            - JSON
                            - WKT
                            - WKB
                            - area
                            - centroid
                            - extent
                            - first_point
                            - hull_rectangle
                            - is_multipart
                            - label_point
                            - last_point
                            - length
                            - length3D
                            - part_count
                            - point_count
                            - true_centroid

                            :Valid Functions:

                            - boundary
                            - buffer
                            - contains
                            - convex_hull
                            - crosses
                            - difference
                            - disjoint
                            - distance_to
                            - equals
                            - generalize
                            - intersect
                            - overlaps
                            - symmetric_difference
                            - touches
                            - union
                            - within

                            Everything else will return None

        -----------------   ------------------------------------------------------
        No Engine           Values will return None by default
        =================   ======================================================



        Required Parameters:
          None
        Optional:
          :param data: panda's dataframe containing attribute information
          :param geometry: list/array/geoseries of arcgis.geometry objects
          :param sr: spatial reference of the dataframe.  This can be the factory
           code, WKT string, arcpy.SpatialReference object, or
           arcgis.SpatailReference object.
          :param gis: passing a gis.GIS object set to Pro will ensure arcpy is
           installed and a full swatch of functionality is available to
           the end user.
    """
    _internal_names = ['_data', '_cacher', '_item_cache', '_cache',
                       'is_copy', '_subtyp', '_index',
                       '_default_kind', '_default_fill_value', '_metadata',
                       '__array_struct__', '__array_interface__']
    _metadata = ['sr', '_geometry_column_name', '_gis']
    _geometry_column_name = GEO_COLUMN_DEFAULT
    #----------------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        """
        A Spatial Dataframe is an object to manipulate, manage and translate
        data into new forms of information for users.

        Required Parameters:
          None

        =====================  ===============================================================
        **optional argument**      **Description**
        ---------------------  ---------------------------------------------------------------
        data                   optional Panda's dataframe, object containing the attribute
                               information.
        ---------------------  ---------------------------------------------------------------
        index                  optional Index or array-like
                               Index to use for resulting frame. Will default to np.arange(n)
                               if no indexing information part of input data and no index
                               provided
        ---------------------  ---------------------------------------------------------------
        columns                optional Index or array-like, Column labels to use for
                               resulting frame. Will default to np.arange(n) if no column
                               labels are provided
        ---------------------  ---------------------------------------------------------------
        dtype                  dytpe, default None, Data type to force, otherwise infer
        ---------------------  ---------------------------------------------------------------
        copy                   optional boolean, default False. Copy data from inputs.
        ---------------------  ---------------------------------------------------------------
        geometry               optional list, default None, list/array/geoseries of
                               arcgis.geometry objects
        ---------------------  ---------------------------------------------------------------
        sr                     optional spatial reference of the dataframe.
        ---------------------  ---------------------------------------------------------------
        gis                    optional gis.GIS object, default None. The GIS object allowes
                               users to use non-public GIS information.
        =====================  ===============================================================

        Example: Creating SpatialDataFrame from a CSV

        df = pd.read_csv(r'D:\ipython_working_folder\joel\store_locations.csv', index_col='OBJECTID')
        geoms = []
        for i in range(0, len(df)):
            x = df.iloc[i]['X']
            y = df.iloc[i]['Y']
            geoms.append(Point({"x" : x, "y" : y, "spatialReference" : {"wkid" : 4326}}))
        sdf = arcgis.features.SpatialDataFrame(data=df, geometry=geoms)

        Example: Creating SpatialDataFrame Using List Comprehension

        coords = [[1,2], [3,4]]
        sdf = SpatialDataFrame(df,
                     geometry=[arcgis.geometry.Geometry({'x':r[0],
                     'y':r[1], 'spatialReference':{'wkid':4326}}) for r in coords])

        .. Note: When passing in a geometry to the SpatialDataFrame, always assign it to the parameter geometry=<var>

        Example: Creating From Feature Class

        sdf = SpatialDataFrame.from_featureclass(r"c:\temp\data.gdb\cities)

        Example: Create A SpatialDataFrame from a Service

        gis = GIS(username="user1", password="password2")
        item = gis.content.search("Roads")[0]
        feature_layer = item.layers[0]
        sdf = SpatialDataFrame.from_layer(feature_layer)

        """
        if not HAS_PANDAS:
            warnings.warn("pandas and numpy are required for SpatialDataFrame.")
            warnings.warn("Please install them.")
        gis = kwargs.pop('gis', arcgis.env.active_gis)
        self._gis = gis
        sr = self._sr(kwargs.pop('sr', 4326))
        geometry = kwargs.pop('geometry', None)
        super(SpatialDataFrame, self).__init__(*args, **kwargs)

        if isinstance(sr, _types.SpatialReference):
            self.sr = sr
        elif isinstance(sr, integer_types):
            self.sr = _types.SpatialReference({'wkid' : sr})
        elif isinstance(sr, string_types):
            self.sr = _types.SpatialReference({'wkt' : sr})
        elif hasattr(sr, 'factoryCode'):
            self.sr = _types.SpatialReference({'wkid' : sr.factoryCode})
        elif hasattr(sr, 'exportToString'):
            self.sr = _types.SpatialReference({'wkt' : sr.exportToString()})
        elif not sr is None:
            raise ValueError("sr (spatial reference) must be a _types.SpatialReference object")
        else:
            self.sr = None
        if geometry is not None:
            self.set_geometry(geometry, inplace=True)
        elif 'SHAPE' in self.columns:
            if isinstance(self['SHAPE'], (GeoSeries, pd.Series)):
                if all(isinstance(x, _types.Geometry) for x in self[self._geometry_column_name]) == False:
                    geometry = [_types.Geometry(g) for g in self['SHAPE'].tolist()]
                    del self['SHAPE']
                    self.set_geometry(geometry, inplace=True)
        if self.sr is None:
            try:
                self.sr = self.geometry[self.geometry.first_valid_index()].spatialReference
            except:
                self.sr = self._sr(sr)
        self._delete_index()
    #----------------------------------------------------------------------
    @property
    def _constructor(self):
        """constructor for class as per Pandas' github page"""
        return SpatialDataFrame
    #----------------------------------------------------------------------
    def info(self, verbose=None, buf=None, max_cols=None, memory_usage=None, null_counts=None):
        """
        Concise summary of a DataFrame.

        Parameters
        ----------
        verbose : {None, True, False}, optional
            Whether to print the full summary.
            None follows the `display.max_info_columns` setting.
            True or False overrides the `display.max_info_columns` setting.
        buf : writable buffer, defaults to sys.stdout
        max_cols : int, default None
            Determines whether full summary or short summary is printed.
            None follows the `display.max_info_columns` setting.
        memory_usage : boolean/string, default None
            Specifies whether total memory usage of the DataFrame
            elements (including index) should be displayed. None follows
            the `display.memory_usage` setting. True or False overrides
            the `display.memory_usage` setting. A value of 'deep' is equivalent
            of True, with deep introspection. Memory usage is shown in
            human-readable units (base-2 representation).
        null_counts : boolean, default None
            Whether to show the non-null counts

            - If None, then only show if the frame is smaller than
              max_info_rows and max_info_columns.
            - If True, always show counts.
            - If False, never show counts.

        """
        cols = list(self.columns)
        cols.pop(cols.index("SHAPE"))
        return self[cols].info(verbose, buf, max_cols,
                               memory_usage, null_counts)
    #----------------------------------------------------------------------
    def _sr(self, sr):
        """sets the spatial reference"""
        if isinstance(sr, _types.SpatialReference):
            return sr
        elif isinstance(sr, dict):
            return _types.SpatialReference(sr)
        elif isinstance(sr, integer_types):
            return _types.SpatialReference({'wkid' : sr})
        elif isinstance(sr, string_types):
            return _types.SpatialReference({'wkt' : sr})
        elif hasattr(sr, 'factoryCode'):
            return _types.SpatialReference({'wkid' : sr.factoryCode})
        elif hasattr(sr, 'exportToString'):
            return _types.SpatialReference({'wkt' : sr.exportToString()})
        elif not sr is None:
            raise ValueError("sr (spatial reference) must be a _types.SpatialReference object")
        else:
            return None
    #----------------------------------------------------------------------
    @property
    def __feature_set__(self):
        """returns a dictionary representation of an Esri FeatureSet"""
        import numpy as np
        import datetime
        import time
        cols_norm = [col for col in self.columns]
        cols_lower = [col.lower() for col in self.columns]
        fields = []
        features = []
        date_fields = []
        _geom_types = {
            arcgis.geometry._types.Point :  "esriGeometryPoint",
            arcgis.geometry._types.Polyline : "esriGeometryPolyline",
            arcgis.geometry._types.MultiPoint : "esriGeometryMultipoint",
            arcgis.geometry._types.Polygon : "esriGeometryPolygon"
        }
        if self.sr is None:
            sr = {'wkid' : 4326}
        else:
            sr = self.sr
        fs = {
            "objectIdFieldName" : "",
            "globalIdFieldName" : "",
            "displayFieldName" : "",
            "geometryType" : _geom_types[type(self.geometry[self.geometry.first_valid_index()])],
            "spatialReference" : sr,
            "fields" : [],
            "features" : []
        }
        if 'objectid' in cols_lower:
            fs['objectIdFieldName'] = cols_norm[cols_lower.index('objectid')]
            fs['displayFieldName'] = cols_norm[cols_lower.index('objectid')]
        elif 'fid' in cols_lower:
            fs['objectIdFieldName'] = cols_norm[cols_lower.index('fid')]
            fs['displayFieldName'] = cols_norm[cols_lower.index('fid')]
        elif 'oid' in cols_lower:
            fs['objectIdFieldName'] = cols_norm[cols_lower.index('oid')]
            fs['displayFieldName'] = cols_norm[cols_lower.index('oid')]
        else:
            self['OBJECTID'] = list(range(1, self.shape[0] + 1))
            res = self.__feature_set__
            del self['OBJECTID']
            return res
        if 'objectIdFieldName' in fs:
            fields.append({
                "name" : fs['objectIdFieldName'],
                "type" : "esriFieldTypeOID",
                "alias" : fs['objectIdFieldName']
            })
            cols_norm.pop(cols_norm.index(fs['objectIdFieldName']))
        if 'globalIdFieldName' in fs and len(fs['globalIdFieldName']) > 0:
            fields.append({
                "name" : fs['globalIdFieldName'],
                "type" : "esriFieldTypeGlobalID",
                "alias" : fs['globalIdFieldName']
            })
            cols_norm.pop(cols_norm.index(fs['globalIdFieldName']))
        elif 'globalIdFieldName' in fs and \
             len(fs['globalIdFieldName']) == 0:
            del fs['globalIdFieldName']
        if self._geometry_column_name in cols_norm:
            cols_norm.pop(cols_norm.index(self._geometry_column_name))
        for col in cols_norm:
            try:
                idx = self[col].first_valid_index()
                col_val = self[col].loc[idx]
            except:
                col_val = ""
            if isinstance(col_val, (str, np.str)):
                l = self[col].str.len().max()
                if str(l) == 'nan':
                    l = 255

                fields.append({
                    "name" : col,
                    "type" : "esriFieldTypeString",
                    "length" : int(l),
                    "alias" : col
                })
                if fs['displayFieldName'] == "":
                    fs['displayFieldName'] = col
            elif isinstance(col_val, (datetime.datetime,
                                      pd.Timestamp,
                                      np.datetime64,
                                      pd.datetime)):
                fields.append({
                    "name" : col,
                    "type" : "esriFieldTypeDate",
                    "alias" : col
                })
                date_fields.append(col)
            elif isinstance(col_val, (np.int32, np.int16, np.int8)):
                fields.append({
                    "name" : col,
                    "type" : "esriFieldTypeSmallInteger",
                    "alias" : col
                })
            elif isinstance(col_val, (int, np.int, np.int64)):
                fields.append({
                    "name" : col,
                    "type" : "esriFieldTypeInteger",
                    "alias" : col
                })
            elif isinstance(col_val, (float, np.float64)):
                fields.append({
                    "name" : col,
                    "type" : "esriFieldTypeDouble",
                    "alias" : col
                })
            elif isinstance(col_val, (np.float32)):
                fields.append({
                    "name" : col,
                    "type" : "esriFieldTypeSingle",
                    "alias" : col
                })
        fs['fields'] = fields
        for row in self.to_dict('records'):
            geom = {}
            if self._geometry_column_name in row:
                geom = row[self._geometry_column_name]
                del row[self._geometry_column_name]
            for f in date_fields:
                try:
                    row[f] = int(row[f].to_pydatetime().timestamp() * 1000)
                except:
                    row[f] = None
            features.append(
                {
                    "geometry" : dict(geom),
                    "attributes" : row
                }
            )
            del row
            del geom
        fs['features'] = features
        return fs
    #----------------------------------------------------------------------
    @property
    def __geo_interface__(self):
        """returns the object as an Feature Collection JSON string"""
        if HASARCPY:
            template = {
                "type": "FeatureCollection",
                "features": []
            }
            geom_type = self.geometry_type
            if geom_type.lower() == "point":
                geom_type = "Point"
            elif geom_type.lower() == "polyline":
                geom_type = "LineString"
            elif geom_type.lower() == "polygon":
                geom_type = "Polygon"
            df_copy = self.copy(deep=True)
            df_copy['geom_json'] = self.geometry.JSON
            df_copy['SHAPE'] = df_copy['geom_json']
            del df_copy['geom_json']
            for index, row in df_copy.iterrows():
                geom = row['SHAPE']
                del row['SHAPE']
                template['features'].append(
                    {"type" : geom_type,
                     "geometry" : pd.io.json.loads(geom),
                     "attributes":row}
                )
            return pd.io.json.dumps(template)
    @property
    def geoextent(self):
        """returns the extent of the spatial dataframe"""
        return self.series_extent
    #----------------------------------------------------------------------
    def __getstate__(self):
        meta = {k: getattr(self, k, None) for k in self._metadata}
        return dict(_data=self._data, _typ=self._typ,
                    _metadata=self._metadata, **meta)
    #----------------------------------------------------------------------
    def __setattr__(self, attr, val):
        if attr.lower() in ['geometry', 'shape', 'shape@']:
            object.__setattr__(self, attr, val)
        else:
            super(SpatialDataFrame, self).__setattr__(attr, val)
    #----------------------------------------------------------------------
    def _get_geometry(self):
        """returns the geometry series"""
        if self._geometry_column_name not in self.columns:
            raise AttributeError("Geometry Column Not Present: %s" % self._geometry_column_name)
        return self[self._geometry_column_name]
    #----------------------------------------------------------------------
    def _set_geometry(self, col):
        """sets the geometry for the panda's dataframe"""
        if isinstance(col, (GeoSeries, list, numpy.array, numpy.ndarray, Series)):
            self.set_geometry(col, inplace=True)
        else:
            raise ValueError("Must be a list, np.array, or GeoSeries")
    #----------------------------------------------------------------------
    geometry = property(fget=_get_geometry,
                        fset=_set_geometry,
                        fdel=None,
                        doc="Get/Set the geometry data for SpatialDataFrame")
    #----------------------------------------------------------------------
    def __finalize__(self, other, method=None, **kwargs):
        """propagate metadata from other to self """
        # merge operation: using metadata of the left object
        if method == 'merge':
            for name in self._metadata:
                object.__setattr__(self, name, getattr(other.left, name, None))
        # concat operation: using metadata of the first object
        elif method == 'concat':
            for name in self._metadata:
                object.__setattr__(self, name, getattr(other.objs[0], name, None))
        else:
            for name in self._metadata:
                object.__setattr__(self, name, getattr(other, name, None))
        return self
    #----------------------------------------------------------------------
    def copy(self, deep=True):
        """
        Make a copy of this SpatialDataFrame object
        Parameters:

        :deep: boolean, default True
               Make a deep copy, i.e. also copy data
        Returns:
         :copy: of SpatialDataFrame
        """
        data = self._data
        if deep:
            data = data.copy()
        return SpatialDataFrame(data, sr=self.sr).__finalize__(self)
    #----------------------------------------------------------------------
    def plot(self, *args, **kwargs):
        """
        Plot draws the data on a web map. The user can describe in simple terms how to
        renderer spatial data using symbol.  To make the process simplier a pallette
        for which colors are drawn from can be used instead of explicit colors.


        ======================  =========================================================
        **Explicit Argument**   **Description**
        ----------------------  ---------------------------------------------------------
        df                      required SpatialDataFrame or GeoSeries. This is the data
                                to map.
        ----------------------  ---------------------------------------------------------
        map_widget              optional WebMap object. This is the map to display the
                                data on.
        ----------------------  ---------------------------------------------------------
        palette                 optional string/dict.  Color mapping.  For simple renderer,
                                just provide a string.  For more robust renderers like
                                unique renderer, a dictionary can be given.
        ----------------------  ---------------------------------------------------------
        renderer_type           optional string.  Determines the type of renderer to use
                                for the provided dataset. The default is 's' which is for
                                simple renderers.

                                Allowed values:

                                + 's' - is a simple renderer that uses one symbol only.
                                + 'u' - unique renderer symbolizes features based on one
                                        or more matching string attributes.
                                + 'c' - A class breaks renderer symbolizes based on the
                                        value of some numeric attribute.
                                + 'h' - heatmap renders point data into a raster
                                        visualization that emphasizes areas of higher
                                        density or weighted values.
        ----------------------  ---------------------------------------------------------
        symbol_type             optional string. This is the type of symbol the user
                                needs to create.  Valid inputs are: simple, picture, text,
                                or carto.  The default is simple.
        ----------------------  ---------------------------------------------------------
        symbol_type             optional string. This is the symbology used by the
                                geometry.  For example 's' for a Line geometry is a solid
                                line. And '-' is a dash line.

                                Allowed symbol types based on geometries:

                                **Point Symbols**

                                 + 'o' - Circle (default)
                                 + '+' - Cross
                                 + 'D' - Diamond
                                 + 's' - Square
                                 + 'x' - X

                                 **Polyline Symbols**

                                 + 's' - Solid (default)
                                 + '-' - Dash
                                 + '-.' - Dash Dot
                                 + '-..' - Dash Dot Dot
                                 + '.' - Dot
                                 + '--' - Long Dash
                                 + '--.' - Long Dash Dot
                                 + 'n' - Null
                                 + 's-' - Short Dash
                                 + 's-.' - Short Dash Dot
                                 + 's-..' - Short Dash Dot Dot
                                 + 's.' - Short Dot

                                 **Polygon Symbols**

                                 + 's' - Solid Fill (default)
                                 + '\' - Backward Diagonal
                                 + '/' - Forward Diagonal
                                 + '|' - Vertical Bar
                                 + '-' - Horizontal Bar
                                 + 'x' - Diagonal Cross
                                 + '+' - Cross

        ----------------------  ---------------------------------------------------------
        col                     optional string/list. Field or fields used for heatmap,
                                class breaks, or unique renderers.
        ----------------------  ---------------------------------------------------------
        pallette                optional string. The color map to draw from in order to
                                visualize the data.  The default pallette is 'jet'. To
                                get a visual representation of the allowed color maps,
                                use the **display_colormaps** method.
        ----------------------  ---------------------------------------------------------
        alpha                   optional float.  This is a value between 0 and 1 with 1
                                being the default value.  The alpha sets the transparancy
                                of the renderer when applicable.
        ======================  =========================================================

        ** Render Syntax **

        The render syntax allows for users to fully customize symbolizing the data.

        ** Simple Renderer**

        A simple renderer is a renderer that uses one symbol only.

        ======================  =========================================================
        **Optional Argument**   **Description**
        ----------------------  ---------------------------------------------------------
        symbol_type             optional string. This is the type of symbol the user
                                needs to create.  Valid inputs are: simple, picture, text,
                                or carto.  The default is simple.
        ----------------------  ---------------------------------------------------------
        symbol_type             optional string. This is the symbology used by the
                                geometry.  For example 's' for a Line geometry is a solid
                                line. And '-' is a dash line.

                                **Point Symbols**

                                + 'o' - Circle (default)
                                + '+' - Cross
                                + 'D' - Diamond
                                + 's' - Square
                                + 'x' - X

                                **Polyline Symbols**

                                + 's' - Solid (default)
                                + '-' - Dash
                                + '-.' - Dash Dot
                                + '-..' - Dash Dot Dot
                                + '.' - Dot
                                + '--' - Long Dash
                                + '--.' - Long Dash Dot
                                + 'n' - Null
                                + 's-' - Short Dash
                                + 's-.' - Short Dash Dot
                                + 's-..' - Short Dash Dot Dot
                                + 's.' - Short Dot

                                **Polygon Symbols**

                                + 's' - Solid Fill (default)
                                + '\' - Backward Diagonal
                                + '/' - Forward Diagonal
                                + '|' - Vertical Bar
                                + '-' - Horizontal Bar
                                + 'x' - Diagonal Cross
                                + '+' - Cross
        ----------------------  ---------------------------------------------------------
        description             Description of the renderer.
        ----------------------  ---------------------------------------------------------
        rotation_expression     A constant value or an expression that derives the angle
                                of rotation based on a feature attribute value. When an
                                attribute name is specified, it's enclosed in square
                                brackets.
        ----------------------  ---------------------------------------------------------
        rotation_type           String value which controls the origin and direction of
                                rotation on point features. If the rotationType is
                                defined as arithmetic, the symbol is rotated from East in
                                a counter-clockwise direction where East is the 0 degree
                                axis. If the rotationType is defined as geographic, the
                                symbol is rotated from North in a clockwise direction
                                where North is the 0 degree axis.

                                Must be one of the following values:

                                + arithmetic
                                + geographic

        ----------------------  ---------------------------------------------------------
        visual_variables        An array of objects used to set rendering properties.
        ======================  =========================================================

        **Heatmap Renderer**

        The HeatmapRenderer renders point data into a raster visualization that emphasizes
        areas of higher density or weighted values.

        ======================  =========================================================
        **Optional Argument**   **Description**
        ----------------------  ---------------------------------------------------------
        blur_radius             The radius (in pixels) of the circle over which the
                                majority of each point's value is spread.
        ----------------------  ---------------------------------------------------------
        field                   This is optional as this renderer can be created if no
                                field is specified. Each feature gets the same
                                value/importance/weight or with a field where each
                                feature is weighted by the field's value.
        ----------------------  ---------------------------------------------------------
        max_intensity           The pixel intensity value which is assigned the final
                                color in the color ramp.
        ----------------------  ---------------------------------------------------------
        min_intensity           The pixel intensity value which is assigned the initial
                                color in the color ramp.
        ----------------------  ---------------------------------------------------------
        ratio                   A number between 0-1. Describes what portion along the
                                gradient the colorStop is added.
        ======================  =========================================================

        **Unique Renderer**

        This renderer symbolizes features based on one or more matching string attributes.

        ======================  =========================================================
        **Optional Argument**   **Description**
        ----------------------  ---------------------------------------------------------
        background_fill_symbol  A symbol used for polygon features as a background if the
                                renderer uses point symbols, e.g. for bivariate types &
                                size rendering. Only applicable to polygon layers.
                                PictureFillSymbols can also be used outside of the Map
                                Viewer for Size and Predominance and Size renderers.
        ----------------------  ---------------------------------------------------------
        default_label           Default label for the default symbol used to draw
                                unspecified values.
        ----------------------  ---------------------------------------------------------
        default_symbol          Symbol used when a value cannot be matched.
        ----------------------  ---------------------------------------------------------
        field1, field2, field3  Attribute field renderer uses to match values.
        ----------------------  ---------------------------------------------------------
        field_delimiter         String inserted between the values if multiple attribute
                                fields are specified.
        ----------------------  ---------------------------------------------------------
        rotation_expression     A constant value or an expression that derives the angle
                                of rotation based on a feature attribute value. When an
                                attribute name is specified, it's enclosed in square
                                brackets. Rotation is set using a visual variable of type
                                rotation info with a specified field or value expression
                                property.
        ----------------------  ---------------------------------------------------------
        rotation_type           String property which controls the origin and direction
                                of rotation. If the rotation type is defined as
                                arithmetic the symbol is rotated from East in a
                                counter-clockwise direction where East is the 0 degree
                                axis. If the rotation type is defined as geographic, the
                                symbol is rotated from North in a clockwise direction
                                where North is the 0 degree axis.
                                Must be one of the following values:

                                + arithmetic
                                + geographic

        ----------------------  ---------------------------------------------------------
        arcade_expression       An Arcade expression evaluating to either a string or a
                                number.
        ----------------------  ---------------------------------------------------------
        arcade_title            The title identifying and describing the associated
                                Arcade expression as defined in the valueExpression
                                property.
        ----------------------  ---------------------------------------------------------
        visual_variables        An array of objects used to set rendering properties.
        ======================  =========================================================

        **Class Breaks Renderer**

        A class breaks renderer symbolizes based on the value of some numeric attribute.

        ======================  =========================================================
        **Optional Argument**   **Description**
        ----------------------  ---------------------------------------------------------
        background_fill_symbol  A symbol used for polygon features as a background if the
                                renderer uses point symbols, e.g. for bivariate types &
                                size rendering. Only applicable to polygon layers.
                                PictureFillSymbols can also be used outside of the Map
                                Viewer for Size and Predominance and Size renderers.
        ----------------------  ---------------------------------------------------------
        default_label           Default label for the default symbol used to draw
                                unspecified values.
        ----------------------  ---------------------------------------------------------
        default_symbol          Symbol used when a value cannot be matched.
        ----------------------  ---------------------------------------------------------
        method                  Determines the classification method that was used to
                                generate class breaks.

                                Must be one of the following values:

                                + esriClassifyDefinedInterval
                                + esriClassifyEqualInterval
                                + esriClassifyGeometricalInterval
                                + esriClassifyNaturalBreaks
                                + esriClassifyQuantile
                                + esriClassifyStandardDeviation
                                + esriClassifyManual

        ----------------------  ---------------------------------------------------------
        field                   Attribute field used for renderer.
        ----------------------  ---------------------------------------------------------
        min_value               The minimum numeric data value needed to begin class
                                breaks.
        ----------------------  ---------------------------------------------------------
        normalization_field     Used when normalizationType is field. The string value
                                indicating the attribute field by which the data value is
                                normalized.
        ----------------------  ---------------------------------------------------------
        normalization_total     Used when normalizationType is percent-of-total, this
                                number property contains the total of all data values.
        ----------------------  ---------------------------------------------------------
        normalization_type      Determine how the data was normalized.

                                Must be one of the following values:

                                + esriNormalizeByField
                                + esriNormalizeByLog
                                + esriNormalizeByPercentOfTotal
        ----------------------  ---------------------------------------------------------
        rotation_expression     A constant value or an expression that derives the angle
                                of rotation based on a feature attribute value. When an
                                attribute name is specified, it's enclosed in square
                                brackets.
        ----------------------  ---------------------------------------------------------
        rotation_type           A string property which controls the origin and direction
                                of rotation. If the rotation_type is defined as
                                arithmetic, the symbol is rotated from East in a
                                couter-clockwise direction where East is the 0 degree
                                axis. If the rotationType is defined as geographic, the
                                symbol is rotated from North in a clockwise direction
                                where North is the 0 degree axis.

                                Must be one of the following values:

                                + arithmetic
                                + geographic

        ----------------------  ---------------------------------------------------------
        arcade_expression       An Arcade expression evaluating to a number.
        ----------------------  ---------------------------------------------------------
        arcade_title            The title identifying and describing the associated
                                Arcade expression as defined in the arcade_expression
                                property.
        ----------------------  ---------------------------------------------------------
        visual_variables        An object used to set rendering options.
        ======================  =========================================================



        ** Symbol Syntax **

        =======================  =========================================================
        **Optional Argument**    **Description**
        -----------------------  ---------------------------------------------------------
        symbol_type              optional string. This is the type of symbol the user
                                 needs to create.  Valid inputs are: simple, picture, text,
                                 or carto.  The default is simple.
        -----------------------  ---------------------------------------------------------
        symbol_type              optional string. This is the symbology used by the
                                 geometry.  For example 's' for a Line geometry is a solid
                                 line. And '-' is a dash line.

                                 **Point Symbols**

                                 + 'o' - Circle (default)
                                 + '+' - Cross
                                 + 'D' - Diamond
                                 + 's' - Square
                                 + 'x' - X

                                 **Polyline Symbols**

                                 + 's' - Solid (default)
                                 + '-' - Dash
                                 + '-.' - Dash Dot
                                 + '-..' - Dash Dot Dot
                                 + '.' - Dot
                                 + '--' - Long Dash
                                 + '--.' - Long Dash Dot
                                 + 'n' - Null
                                 + 's-' - Short Dash
                                 + 's-.' - Short Dash Dot
                                 + 's-..' - Short Dash Dot Dot
                                 + 's.' - Short Dot

                                 **Polygon Symbols**

                                 + 's' - Solid Fill (default)
                                 + '\' - Backward Diagonal
                                 + '/' - Forward Diagonal
                                 + '|' - Vertical Bar
                                 + '-' - Horizontal Bar
                                 + 'x' - Diagonal Cross
                                 + '+' - Cross
        -----------------------  ---------------------------------------------------------
        cmap                     optional string or list.  This is the color scheme a user
                                 can provide if the exact color is not needed, or a user
                                 can provide a list with the color defined as:
                                 [red, green blue, alpha]. The values red, green, blue are
                                 from 0-255 and alpha is a float value from 0 - 1.
                                 The default value is 'jet' color scheme.
        -----------------------  ---------------------------------------------------------
        cstep                    optional integer.  If provided, its the color location on
                                 the color scheme.
        =======================  =========================================================

        **Simple Symbols**

        This is a list of optional parameters that can be given for point, line or
        polygon geometries.

        ====================  =========================================================
        **Argument**          **Description**
        --------------------  ---------------------------------------------------------
        marker_size           optional float.  Numeric size of the symbol given in
                              points.
        --------------------  ---------------------------------------------------------
        marker_angle          optional float. Numeric value used to rotate the symbol.
                              The symbol is rotated counter-clockwise. For example,
                              The following, angle=-30, in will create a symbol rotated
                              -30 degrees counter-clockwise; that is, 30 degrees
                              clockwise.
        --------------------  ---------------------------------------------------------
        marker_xoffset        Numeric value indicating the offset on the x-axis in points.
        --------------------  ---------------------------------------------------------
        marker_yoffset        Numeric value indicating the offset on the y-axis in points.
        --------------------  ---------------------------------------------------------
        line_width            optional float. Numeric value indicating the width of the line in points
        --------------------  ---------------------------------------------------------
        outline_style         Optional string. For polygon point, and line geometries , a
                              customized outline type can be provided.

                              Allowed Styles:

                              + 's' - Solid (default)
                              + '-' - Dash
                              + '-.' - Dash Dot
                              + '-..' - Dash Dot Dot
                              + '.' - Dot
                              + '--' - Long Dash
                              + '--.' - Long Dash Dot
                              + 'n' - Null
                              + 's-' - Short Dash
                              + 's-.' - Short Dash Dot
                              + 's-..' - Short Dash Dot Dot
                              + 's.' - Short Dot
        --------------------  ---------------------------------------------------------
        outline_color         optional string or list.  This is the same color as the
                              cmap property, but specifically applies to the outline_color.
        ====================  =========================================================

        **Picture Symbol**

        This type of symbol only applies to Points, MultiPoints and Polygons.

        ====================  =========================================================
        **Argument**          **Description**
        --------------------  ---------------------------------------------------------
        marker_angle          Numeric value that defines the number of degrees ranging
                              from 0-360, that a marker symbol is rotated. The rotation
                              is from East in a counter-clockwise direction where East
                              is the 0 axis.
        --------------------  ---------------------------------------------------------
        marker_xoffset        Numeric value indicating the offset on the x-axis in points.
        --------------------  ---------------------------------------------------------
        marker_yoffset        Numeric value indicating the offset on the y-axis in points.
        --------------------  ---------------------------------------------------------
        height                Numeric value used if needing to resize the symbol. Specify a value in points. If images are to be displayed in their original size, leave this blank.
        --------------------  ---------------------------------------------------------
        width                 Numeric value used if needing to resize the symbol. Specify a value in points. If images are to be displayed in their original size, leave this blank.
        --------------------  ---------------------------------------------------------
        url                   String value indicating the URL of the image. The URL should be relative if working with static layers. A full URL should be used for map service dynamic layers. A relative URL can be dereferenced by accessing the map layer image resource or the feature layer image resource.
        --------------------  ---------------------------------------------------------
        image_data            String value indicating the base64 encoded data.
        --------------------  ---------------------------------------------------------
        xscale                Numeric value indicating the scale factor in x direction.
        --------------------  ---------------------------------------------------------
        yscale                Numeric value indicating the scale factor in y direction.
        --------------------  ---------------------------------------------------------
        outline_color         optional string or list.  This is the same color as the
                              cmap property, but specifically applies to the outline_color.
        --------------------  ---------------------------------------------------------
        outline_style         Optional string. For polygon point, and line geometries , a
                              customized outline type can be provided.

                              Allowed Styles:

                              + 's' - Solid (default)
                              + '-' - Dash
                              + '-.' - Dash Dot
                              + '-..' - Dash Dot Dot
                              + '.' - Dot
                              + '--' - Long Dash
                              + '--.' - Long Dash Dot
                              + 'n' - Null
                              + 's-' - Short Dash
                              + 's-.' - Short Dash Dot
                              + 's-..' - Short Dash Dot Dot
                              + 's.' - Short Dot
        --------------------  ---------------------------------------------------------
        outline_color         optional string or list.  This is the same color as the
                              cmap property, but specifically applies to the outline_color.
        --------------------  ---------------------------------------------------------
        line_width            optional float. Numeric value indicating the width of the line in points
        ====================  =========================================================

        **Text Symbol**

        This type of symbol only applies to Points, MultiPoints and Polygons.

        ====================  =========================================================
        **Argument**          **Description**
        --------------------  ---------------------------------------------------------
        font_decoration       The text decoration. Must be one of the following values:
                              - line-through
                              - underline
                              - none
        --------------------  ---------------------------------------------------------
        font_family           Optional string. The font family.
        --------------------  ---------------------------------------------------------
        font_size             Optional float. The font size in points.
        --------------------  ---------------------------------------------------------
        font_style            Optional string. The text style.
                              - italic
                              - normal
                              - oblique
        --------------------  ---------------------------------------------------------
        font_weight           Optional string. The text weight.
                              Must be one of the following values:
                              - bold
                              - bolder
                              - lighter
                              - normal
        --------------------  ---------------------------------------------------------
        background_color      optional string/list. Background color is represented as
                              a four-element array or string of a color map.
        --------------------  ---------------------------------------------------------
        halo_color            Optional string/list. Color of the halo around the text.
                              The default is None.
        --------------------  ---------------------------------------------------------
        halo_size             Optional integer/float. The point size of a halo around
                              the text symbol.
        --------------------  ---------------------------------------------------------
        horizontal_alignment  optional string. One of the following string values
                              representing the horizontal alignment of the text.
                              Must be one of the following values:
                              - left
                              - right
                              - center
                              - justify
        --------------------  ---------------------------------------------------------
        kerning               optional boolean. Boolean value indicating whether to
                              adjust the spacing between characters in the text string.
        --------------------  ---------------------------------------------------------
        line_color            optional string/list. Outline color is represented as
                              a four-element array or string of a color map.
        --------------------  ---------------------------------------------------------
        line_width            optional integer/float. Outline size.
        --------------------  ---------------------------------------------------------
        marker_angle          optional int. A numeric value that defines the number of
                              degrees (0 to 360) that a text symbol is rotated. The
                              rotation is from East in a counter-clockwise direction
                              where East is the 0 axis.
        --------------------  ---------------------------------------------------------
        marker_xoffset        optional int/float.Numeric value indicating the offset
                              on the x-axis in points.
        --------------------  ---------------------------------------------------------
        marker_yoffset        optional int/float.Numeric value indicating the offset
                              on the x-axis in points.
        --------------------  ---------------------------------------------------------
        right_to_left         optional boolean. Set to true if using Hebrew or Arabic
                              fonts.
        --------------------  ---------------------------------------------------------
        rotated               optional boolean. Boolean value indicating whether every
                              character in the text string is rotated.
        --------------------  ---------------------------------------------------------
        text                  Required string.  Text Value to display next to geometry.
        --------------------  ---------------------------------------------------------
        vertical_alignment    Optional string. One of the following string values
                              representing the vertical alignment of the text.
                              Must be one of the following values:
                              - top
                              - bottom
                              - middle
                              - baseline
        ====================  =========================================================

        **Cartographic Symbol**

        This type of symbol only applies to line geometries.

        ====================  =========================================================
        **Argument**          **Description**
        --------------------  ---------------------------------------------------------
        line_width            optional float. Numeric value indicating the width of the line in points
        --------------------  ---------------------------------------------------------
        cap                   Optional string.  The cap style.
        --------------------  ---------------------------------------------------------
        join                  Optional string. The join style.
        --------------------  ---------------------------------------------------------
        miter_limit           Optional string. Size threshold for showing mitered line joins.
        ====================  =========================================================

        The kwargs parameter accepts all parameters of the create_symbol method and the
        create_renderer method.


        """
        if ('kind' in kwargs and \
           kwargs['kind'] == 'map') or \
           (len(args) > 3 and args[3] == 'map'):
            from arcgis.features._data.geodataset.viz import plot
            has_wm = True
            wm = kwargs.pop('map_widget', None)
            if wm is None:
                has_wm = False
                wm = GIS().map()
            if has_wm:
                plot(df=self,
                     map_widget=wm,
                     name=kwargs.pop('name', "Feature Collection Layer"),
                     renderer_type=kwargs.pop("renderer_type", None),
                     symbol_type=kwargs.pop('symbol_type', None),
                     symbol_style=kwargs.pop('symbol_style', None),
                     col=kwargs.pop('col', None),
                     colors=kwargs.pop('cmap', None) or kwargs.pop('colors', None) or kwargs.pop('pallette', 'jet'),
                     alpha=kwargs.pop('alpha', 1),
                     **kwargs)
                return True
            else:
                return plot(df=self,
                            map_widget=wm,
                            name=kwargs.pop('name', "Feature Collection Layer"),
                            renderer_type=kwargs.pop("renderer_type", None),
                            symbol_type=kwargs.pop('symbol_type', None),
                            symbol_style=kwargs.pop('symbol_style', None),
                            col=kwargs.pop('col', None),
                            colors=kwargs.pop('cmap', None) or kwargs.pop('colors', None) or kwargs.pop('pallette', 'jet'),
                            alpha=kwargs.pop('alpha', 1),
                            **kwargs)
        if ('kind' in kwargs and \
           kwargs['kind'] == 'map') or \
           (len(args) > 3 and args[3] == 'map') and \
           ('as_graphic' in kwargs and kwargs['as_graphic']):
            from arcgis.features import FeatureCollection, FeatureSet
            from arcgis import geometry
            if self._gis is None:
                gis = GIS(set_active=False)
            else:
                gis = self._gis
            if self.sr:
                sr = self.sr
            else:
                sr = self.sr
            extent = None
            if HASARCPY:
                if sr:
                    wkid = None
                    if hasattr(sr, 'factoryCode'):
                        wkid = {'wkid' : sr.factoryCode}
                    elif isinstance(sr, geometry.SpatialReference):
                        wkid = self.sr
                    ext = self.geoextent
                    extent = {
                        "xmin" : ext[0],
                        "ymin" : ext[1],
                        "xmax" : ext[2],
                        "ymax" : ext[3],
                        "spatialReference" : wkid
                    }
                else:
                    ext = self.geoextent
                    extent = {
                        "xmin" : ext[0],
                        "ymin" : ext[1],
                        "xmax" : ext[2],
                        "ymax" : ext[3],
                        "spatialReference" : {'wkid' : 4326}
                    }
            else:
                sr = self.sr
                if self.sr is None:
                    sr = {'wkid' : 4326}

                ext = self.geoextent
                extent = {
                    "xmin" : ext[0],
                    "ymin" : ext[1],
                    "xmax" : ext[2],
                    "ymax" : ext[3],
                    "spatialReference" : sr
                }
            if 'map_widget' not in kwargs:
                raise Exception("map_widget is required to plot the SpatialDataFrame")
            else:
                m = kwargs.pop('map_widget')
                symbol = kwargs.pop('symbol', None)
                popup = kwargs.pop('popup', None)
            try:
                fs = FeatureSet.from_dict(self.__feature_set__)
                m.draw(fs, symbol=symbol, popup=popup)
                if extent and \
                   isinstance(extent, dict):
                    m.extent = extent
            except:
                raise Exception('Could not plot the Spatial DataFrame.')
        else:
            return super(SpatialDataFrame, self).plot(*args, **kwargs)
    # ----------------------------------------------------------------------
    @staticmethod
    def from_xy(df, x_column, y_column, sr=4326):
        """
        Converts a Pandas DataFrame into a Spatial DataFrame by providing the X/Y columns.

        ====================    =========================================================
        **Argument**            **Description**
        --------------------    ---------------------------------------------------------
        df                      Required Pandas DataFrame. Source dataset
        --------------------    ---------------------------------------------------------
        x_column                Required string.  The name of the X-coordinate series
        --------------------    ---------------------------------------------------------
        y_column                Required string.  The name of the Y-coordinate series
        --------------------    ---------------------------------------------------------
        sr                      Optional int.  The wkid number of the spatial reference.
        ====================    =========================================================

        :returns: SpatialDataFrame

        """
        from .io.fileops import _from_xy
        return _from_xy(df=df, x_column=x_column,
                        y_column=y_column, sr=sr)
    # ----------------------------------------------------------------------
    @staticmethod
    def from_df(df, address_column="address", geocoder=None):
        """
        Returns a SpatialDataFrame from a dataframe with an address column.

        ====================    =========================================================
        **Argument**            **Description**
        --------------------    ---------------------------------------------------------
        df                      Required Pandas DataFrame. Source dataset
        --------------------    ---------------------------------------------------------
        address_column          Optional String. The default is "address". This is the
                                name of a column in the specified dataframe that contains
                                addresses (as strings). The addresses are batch geocoded
                                using the GIS's first configured geocoder and their
                                locations used as the geometry of the spatial dataframe.
                                Ignored if the 'geometry' parameter is also specified.
        --------------------    ---------------------------------------------------------
        geocoder                Optional Geocoder. The geocoder to be used. If not
                                specified, the active GIS's first geocoder is used.
        ====================    =========================================================

        :returns: SpatialDataFrame



        NOTE: Credits will be consumed for batch_geocoding, from
        the GIS to which the geocoder belongs.

        """
        from arcgis.geocoding import get_geocoders, geocode, batch_geocode
        if geocoder is None:
            geocoder = arcgis.env.active_gis._tools.geocoders[0]

        geoms = []
        if address_column in df.columns:
            # batch geocode addresses in the address column and use them as the geometry
            batch_size = geocoder.properties.locatorProperties.MaxBatchSize
            N = len(df)
            geoms = []
            for i in range(0, N, batch_size):
                start = i
                stop = i + batch_size if i + batch_size < N else N
                # print('Geocoding from ' + str(start) + ' to ' + str(stop))

                res = batch_geocode(list(df[start:stop][address_column]), geocoder=geocoder)
                for index in range(len(res)):
                    address = df.ix[start + index, address_column]
                    try:
                        loc = res[index]['location']
                        x = loc['x']
                        y = loc['y']
                        # self.ix[start + index, 'x'] = x
                        # self.ix[start + index, 'y'] = y
                        geoms.append(arcgis.geometry.Geometry({'x': x, 'y': y}))

                    except:
                        x, y = None, None
                        try:
                            loc = geocode(address, geocoder=geocoder)[0]['location']
                            x = loc['x']
                            y = loc['y']
                        except:
                            print('Unable to geocode address: ' + address)
                            pass
                        # self.ix[start + index, 'x'] = x
                        # self.ix[start + index, 'y'] = y
                        geoms.append(None)
        else:
            raise ValueError("Address column not found in dataframe")

        return SpatialDataFrame(df, geometry=geoms)
    #----------------------------------------------------------------------
    @staticmethod
    def from_featureclass(filename, **kwargs):
        """
        Returns a SpatialDataFrame from a feature class.

        ====================    =========================================================
        **Argument**            **Description**
        --------------------    ---------------------------------------------------------
        filename                Required string. The full path to the feature class
        --------------------    ---------------------------------------------------------
        sql_clause              Optional string. The sql clause to parse data down
        --------------------    ---------------------------------------------------------
        where_clause            Optional string. A where statement
        --------------------    ---------------------------------------------------------
        sr                      Optional SpatialReference. A spatial reference object
        ====================    =========================================================

        :returns: SpatialDataFrame

        """
        from .io import from_featureclass
        gis = kwargs.pop('gis', arcgis.env.active_gis)
        if HASARCPY:
            return from_featureclass(filename=filename, **kwargs)
        elif isinstance(gis, GIS) and \
             gis._con._auth.lower() != "anon":
            return from_featureclass(filename=filename, **kwargs)
        else:
            raise Exception("Cannot create the SpatialDataFrame, you must " +\
                            "have an authenticated GIS.")
    #----------------------------------------------------------------------
    @staticmethod
    def from_layer(layer, **kwargs):
        """
        Returns a SpatialDataFrame/Pandas' Dataframe from a FeatureLayer or Table object.

        ==============   ==============================================================
        **Arguments**    **Description**
        --------------   --------------------------------------------------------------
        layer            required FeatureLayer/Table. This is the service endpoint object.
        ==============   ==============================================================

        :returns: SpatialDataFrame for feature layers with geometry and Panda's Dataframe for tables

        """
        from .io import from_layer
        return from_layer(layer=layer, **kwargs)
    #----------------------------------------------------------------------
    def to_featureclass(self,
                        out_location, out_name,
                        overwrite=True, skip_invalid=True):
        """converts a SpatialDataFrame to a feature class

        ====================    =========================================================
        **Argument**            **Description**
        --------------------    ---------------------------------------------------------
        out_location            Required string. A save location workspace
        --------------------    ---------------------------------------------------------
        out_name                Required string. The name of the feature class to save as
        --------------------    ---------------------------------------------------------
        overwrite               Optional boolean. True means to erase and replace value,
                                false means to append
        --------------------    ---------------------------------------------------------
        skip_invalids           Optional boolean. If True, any bad rows will be ignored.
        ====================    =========================================================

        :returns: string

        """
        from .io import to_featureclass
        return to_featureclass(df=self,
                               out_location=out_location,
                               out_name=out_name,
                               overwrite=overwrite, skip_invalid=skip_invalid)
    #----------------------------------------------------------------------
    def to_hdf(self, path_or_buf, key, **kwargs):
        """Write the contained data to an HDF5 file using HDFStore.

        Parameters
        ----------
        path_or_buf : the path (string) or HDFStore object
        key : string
            indentifier for the group in the store
        mode : optional, {'a', 'w', 'r+'}, default 'a'

          ``'w'``
              Write; a new file is created (an existing file with the same
              name would be deleted).
          ``'a'``
              Append; an existing file is opened for reading and writing,
              and if the file does not exist it is created.
          ``'r+'``
              It is similar to ``'a'``, but the file must already exist.
        format : 'fixed(f)|table(t)', default is 'fixed'
            fixed(f) : Fixed format
                       Fast writing/reading. Not-appendable, nor searchable
            table(t) : Table format
                       Write as a PyTables Table structure which may perform
                       worse but allow more flexible operations like searching
                       / selecting subsets of the data
        append : boolean, default False
            For Table formats, append the input data to the existing
        data_columns :  list of columns, or True, default None
            List of columns to create as indexed data columns for on-disk
            queries, or True to use all columns. By default only the axes
            of the object are indexed. See `here
            <http://pandas.pydata.org/pandas-docs/stable/io.html#query-via-data-columns>`__.

            Applicable only to format='table'.
        complevel : int, 1-9, default 0
            If a complib is specified compression will be applied
            where possible
        complib : {'zlib', 'bzip2', 'lzo', 'blosc', None}, default None
            If complevel is > 0 apply compression to objects written
            in the store wherever possible
        fletcher32 : bool, default False
            If applying compression use the fletcher32 checksum
        dropna : boolean, default False.
            If true, ALL nan rows will not be written to store.
        """

        from pandas.io import pytables
        return pytables.to_hdf(path_or_buf, key, pd.DataFrame(self), **kwargs)
    #----------------------------------------------------------------------
    @staticmethod
    def from_hdf(path_or_buf, key=None, **kwargs):
        """ read from the store, close it if we opened it

            Retrieve pandas object stored in file, optionally based on where
            criteria

            Parameters
            ----------
            path_or_buf : path (string), buffer, or path object (pathlib.Path or
                py._path.local.LocalPath) to read from

                .. versionadded:: 0.19.0 support for pathlib, py.path.

            key : group identifier in the store. Can be omitted if the HDF file
                contains a single pandas object.
            where : list of Term (or convertable) objects, optional
            start : optional, integer (defaults to None), row number to start
                selection
            stop  : optional, integer (defaults to None), row number to stop
                selection
            columns : optional, a list of columns that if not None, will limit the
                return columns
            iterator : optional, boolean, return an iterator, default False
            chunksize : optional, nrows to include in iteration, return an iterator

            Returns
            -------
            The selected object

            """
        return SpatialDataFrame(pd.read_hdf(path_or_buf=path_or_buf,
                                            key=key, **kwargs))
    #----------------------------------------------------------------------
    def to_feature_collection(self,
                              name=None,
                              drawing_info=None,
                              extent=None,
                              global_id_field=None):
        """
        converts a Spatial DataFrame to a Feature Collection

        =====================  ===============================================================
        **optional argument**  **Description**
        ---------------------  ---------------------------------------------------------------
        name                   optional string. Name of the Feature Collection
        ---------------------  ---------------------------------------------------------------
        drawing_info           Optional dictionary. This is the rendering information for a
                               Feature Collection.  Rendering information is a dictionary with
                               the symbology, labelling and other properties defined.  See:
                               http://resources.arcgis.com/en/help/arcgis-rest-api/index.html#/Renderer_objects/02r30000019t000000/
        ---------------------  ---------------------------------------------------------------
        extent                 Optional dictionary.  If desired, a custom extent can be
                               provided to set where the map starts up when showing the data.
                               The default is the full extent of the dataset in the Spatial
                               DataFrame.
        ---------------------  ---------------------------------------------------------------
        global_id_field        Optional string. The Global ID field of the dataset.
        =====================  ===============================================================

        :returns: FeatureCollection object
        """
        from arcgis.features import FeatureCollection
        import uuid
        import string
        import random

        if name is None:
            name = random.choice(string.ascii_letters) + uuid.uuid4().hex[:5]
        template = {
            'showLegend' : True,
            'layers' : []
        }
        if extent is None:
            ext = self.geoextent
            extent = {
                "xmin" : ext[0],
                "ymin" : ext[1],
                "xmax" : ext[2],
                "ymax" : ext[3],
                "spatialReference" : self.sr
            }
        fs = self.__feature_set__
        fields = []
        for fld in fs['fields']:
            if fld['name'].lower() == fs['objectIdFieldName'].lower():
                fld['editable'] = False
                fld['sqlType'] = "sqlTypeOther"
                fld['domain'] = None
                fld['defaultValue'] = None
                fld['nullable'] = False
            else:
                fld['editable'] = True
                fld['sqlType'] = "sqlTypeOther"
                fld['domain'] = None
                fld['defaultValue'] = None
                fld['nullable'] = True
        if drawing_info is None:
            di = {
                'renderer' : {
                    'labelingInfo' : None,
                    'label' : "",
                    'description' : "",
                    'type' : 'simple',
                    'symbol' : None

                }
            }
            symbol = None
            if symbol is None:
                if fs['geometryType'] in ["esriGeometryPoint", "esriGeometryMultipoint"]:
                    di['renderer']['symbol'] = {"color":[0,128,0,128],"size":18,"angle":0,
                                                "xoffset":0,"yoffset":0,
                                                "type":"esriSMS",
                                                "style":"esriSMSCircle",
                                                "outline":{"color":[0,128,0,255],"width":1,
                                                           "type":"esriSLS","style":"esriSLSSolid"}}
                elif fs['geometryType'] == 'esriGeometryPolyline':
                    di['renderer']['symbol'] = {
                        "type": "esriSLS",
                        "style": "esriSLSDot",
                        "color": [0,128,0,128],
                        "width": 1
                    }
                elif fs['geometryType'] == 'esriGeometryPolygon':
                    di['renderer']['symbol'] = {
                        "type": "esriSFS",
                        "style": "esriSFSSolid",
                        "color": [0,128,0,128],
                        "outline": {
                            "type": "esriSLS",
                            "style": "esriSLSSolid",
                            "color": [110,110,110,255],
                            "width": 1
                        }
                    }
            else:
                di['renderer']['symbol'] = symbol
        else:
            di = drawing_info
        layer = {
            'featureSet' : {'features' : fs['features'],
                            'geometryType' : fs['geometryType']
                            },
            'layerDefinition' : {
                'htmlPopupType' : 'esriServerHTMLPopupTypeNone',
                'objectIdField' : fs['objectIdFieldName'] or "OBJECTID",
                #'types' : [],
                'defaultVisibility' : True,
                'supportsValidateSql' : True,
                'supportsAttachmentsByUploadId' : True,
                'useStandardizedQueries' : False,
                'supportsApplyEditsWithGlobalIds' : True,
                'standardMaxRecordCount' : 32000,
                'supportsTruncate' : False,
                'extent' : extent,
                'maxScale' : 0,
                'supportsAppend' : True,
                'supportsCalculate' : True,
                'copyrightText' : "",
                #'templates' : [],
                'description' : "",
                #'relationships' : [],
                'supportsRollbackOnFailureParameter' : True,
                'hasM' : False,
                'displayField' : "",
                'drawingInfo' : di,
                'type' : 'Feature Layer',
                'supportedQueryFormats' : 'JSON, geoJSON',
                'isDataVersioned' : False,
                'maxRecordCount' : 2000,
                'minScale' : 0,
                'supportsStatistics' : True,
                'hasAttachments' : False,
                #'indexes' : [],
                'tileMaxRecordCount' : 8000,
                'supportsAdvancedQueries' : True,
                #'globalIdField' : "",
                'hasZ' : False,
                'name' : name,
                'id' : 0,
                'allowGeometryUpdates' : True,
                #'typeIdField' : "",
                'geometryType' : fs['geometryType'],
                'currentVersion' : 10.51,
                #'maxRecordCountFactor' : 1,
                'supportsCoordinatesQuantization' : True,
                'fields' : fs['fields'],
                'hasStaticData' : True,# False
                'capabilities' : 'Create,Delete,Query,Update,Editing,Extract,Sync',
                'advancedQueryCapabilities' :  {'supportsReturningGeometryCentroid': False,
                                                'supportsQueryRelatedPagination': True,
                                                'supportsHavingClause': True,
                                                'supportsOrderBy': True,
                                                'supportsPaginationOnAggregatedQueries': True,
                                                'supportsQueryWithDatumTransformation': True,
                                                'supportsAdvancedQueryRelated': True,
                                                'supportsOutFieldSQLExpression': True,
                                                'supportsPagination': True,
                                                'supportsStatistics': True,
                                                'supportsSqlExpression': True,
                                                'supportsQueryWithDistance': True,
                                                'supportsReturningQueryExtent': True,
                                                'supportsDistinct': True,
                                                'supportsQueryWithResultType': True},

            }
        }
        if global_id_field is not None:
            layer['layerDefinition']['globalIdField'] = global_id_field
        return FeatureCollection(layer)
    #----------------------------------------------------------------------
    def to_featureset(self):
        """
        Converts a spatial dataframe to a feature set object
        """
        from arcgis.features import FeatureSet
        return FeatureSet.from_dataframe(self)
    #----------------------------------------------------------------------
    def _to_arcpy_featureset(self):
        """
        Converts a Spatial DataFrame to arcpy.FeatureSet so it can be used
        in geoprocessing tools.

        returns: arcpy.FeatureSet
        """
        if HAS_ARCPY:
            import uuid, string, random
            l = []
            for i in range(3):
                l.append(random.choice(string.ascii_letters))
            l = "".join(l)
            out_name = l
            res = self.to_featureclass(out_location='in_memory',
                                       out_name=out_name)

            feature_set = arcpy.FeatureSet()
            feature_set.load(res)
            return feature_set
        else:
            raise Exception("ArcPy must be present to convert to arcpy.FeatureSet object")
    #----------------------------------------------------------------------
    def to_featurelayer(self,
                        title,
                        gis=None,
                        tags=None):
        """
        publishes a spatial dataframe to a new feature layer

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        title                           Required string. The name of the service
        ---------------------------     --------------------------------------------------------------------
        gis                             Optional GIS. The GIS connection object
        ---------------------------     --------------------------------------------------------------------
        tags                            Optional string. A comma seperated list of descriptive words for the
                                        service
        ===========================     ====================================================================

        :returns: FeatureLayer

        """
        from arcgis import env
        if gis is None:
            gis = env.active_gis
            if gis is None:
                raise ValueError("GIS object must be provided")
        content = gis.content
        return content.import_data(self, title=title, tags=tags)
    #----------------------------------------------------------------------
    def set_geometry(self, col, drop=False, inplace=False, sr=None):
        """
        Set the SpatialDataFrame geometry using either an existing column or
        the specified input. By default yields a new object.

        The original geometry column is replaced with the input.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        col                             Required string/np.array. column label or array
        ---------------------------     --------------------------------------------------------------------
        drop                            Optional boolean. Default True. Delete column to be used as the new
                                        geometry
        ---------------------------     --------------------------------------------------------------------
        inplace                         Optional boolean. Default False. Modify the SpatialDataFrame in
                                        place (do not create a new object)
        ---------------------------     --------------------------------------------------------------------
        sr                              Optional SpatialReference/Integer. The wkid value Coordinate
                                        system to use. If passed, overrides both DataFrame and col's sr.
                                        Otherwise, tries to get sr from passed col values or
                                        DataFrame.
        ===========================     ====================================================================

        :returns: SpatialDataFrame

        """
        if inplace:
            frame = self
        else:
            frame = self.copy()
        if sr:
            sr = self._sr(sr=sr)
        if not sr:
            sr = getattr(col, 'sr', None)
            if sr is None and \
               isinstance(col, GeoSeries):
                col.sr = self.sr
        to_remove = None
        if isinstance(col, string_types):
            geo_column_name = col
            self._geometry_column_name = col
        else:
            geo_column_name = self._geometry_column_name
        if isinstance(col, (GeoSeries, Series, list, numpy.ndarray)):
            level = col
        elif hasattr(col, 'ndim') and col.ndim != 1:
            raise ValueError("Must pass array with one dimension only.")
        else:
            try:
                level = frame[col].values
            except KeyError:
                raise ValueError("Unknown column %s" % col)
            except:
                raise
            if drop:
                to_remove = col
                geo_column_name = self._geometry_column_name
            else:
                geo_column_name = col

        if to_remove:
            del frame[to_remove]

        if isinstance(level, GeoSeries) and level.sr != sr:
            # Avoids caching issues/sr sharing issues
            level = level.copy()
            level.sr = sr
        # Check that we are using a listlike of geometries
        if not all(isinstance(item, GEOM_TYPES) or not item for item in level):
            raise TypeError("Input geometry column must contain valid geometry objects.")
        #if isinstance(frame[geo_column_name], pd.Series):
        #    frame[geo_column_name] = GeoSeries(frame[geo_column_name])
        if isinstance(level, (list, tuple, numpy.ndarray)):
            level = GeoSeries(level)
        frame[geo_column_name] = level
        frame._geometry_column_name = geo_column_name
        frame.sr = sr
        frame._delete_index()

        if (frame.sr != self.sr and HASARCPY):
            if isinstance(sr, dict):
                if hasattr(sr, 'as_arcpy') and HASARCPY:
                    sr = sr.as_arcpy
                elif 'wkid' in sr:
                    sr = sr['wkid']
                elif 'wkt' in sr:
                    sr = sr['wkt']
        import json
        gtypes = frame.geometry.apply(lambda x: type(x)).unique()
        if len(gtypes) == 1 and \
           gtypes[0] in [_types.Point, _types.Polygon, _types.Polyline]:
            pass
        elif HASARCPY: # Use ArcPy to Enforce Proper Geometry Construction
            for idx, g in frame.geometry.iteritems():
                if isinstance(g, arcpy.Point):
                    g = arcgis.geometry.Geometry(json.loads(arcpy.PointGeometry(g, sr).JSON))
                elif hasattr(g, "JSON"):
                    g = arcgis.geometry.Geometry(json.loads(g.JSON))
                elif isinstance(g, string_types):
                    g = arcgis.geometry.Geometry(json.loads(g))
                elif isinstance(g, dict):
                    g = arcgis.geometry.Geometry(g)

                if inplace:
                    try:
                        frame.loc[idx, self._geometry_column_name] = g
                    except:
                        try:
                            frame.at[idx, self._geometry_column_name] = g
                        except:
                            frame.set_value(index=idx,
                                        col=self._geometry_column_name,
                                        value=g)
                else:
                    try:
                        frame.iloc[idx, self._geometry_column_name] = g
                    except:
                        frame.loc[idx, self._geometry_column_name] = g
                del idx, g
            if sr:
                frame.sr = self._sr(sr)
                frame.geometry = frame.geometry.project_as(sr)
        else:
            sr = self.sr
            if sr is None:
                sr = {'wkid' : 4326}
            for idx, g in frame.geometry.iteritems():
                if hasattr(g, "JSON") and HASARCPY:
                    g = arcgis.geometry.Geometry(json.loads(g.JSON))
                elif str(type(g)) == "<class 'arcpy.arcobjects.arcobjects.Point'>":
                    g = arcgis.geometry.Geometry({'x':g.X,
                                                  'y':g.Y,
                                                  'spatialReference':sr})
                elif isinstance(g, string_types):
                    g = arcgis.geometry.Geometry(json.loads(g))
                elif isinstance(g, dict):
                    g = arcgis.geometry.Geometry(g)
                else:
                    raise ValueError("Invalid Geometry")
                if sr is None:
                    sr = {'wkid' : 4326}
                if 'spatialReference' not in g:
                    g['spatialReference'] = dict(sr)
                if inplace:
                    try:
                        frame.loc[idx, self._geometry_column_name] = g
                    except:
                        frame.set_value(index=idx, col=self._geometry_column_name, value=g)
                else:
                    try:
                        frame.iloc[idx, self._geometry_column_name] = g
                    except:
                        frame.set_value(index=idx, col=self._geometry_column_name, value=g)
            frame.sr = self._sr(sr)
        if not inplace:
            return frame
        self = frame
    #----------------------------------------------------------------------
    def __getitem__(self, key):
        """
        If the result is a column containing only 'geometry', return a
        GeoSeries. If it's a DataFrame with a 'geometry' column, return a
        SpatialDataFrame.
        """
        result = super(SpatialDataFrame, self).__getitem__(key)
        geo_col = self._geometry_column_name
        if isinstance(key, string_types) and key == geo_col:
            result.__class__ = GeoSeries
            result.sr = self.sr
            result._delete_index()
        elif isinstance(result, DataFrame) and geo_col in result:
            result.__class__ = SpatialDataFrame
            result.sr = self.sr
            result._geometry_column_name = geo_col
            result._delete_index()
        elif isinstance(result, DataFrame) and geo_col not in result:
            result.__class__ = DataFrame
        return result
    #----------------------------------------------------------------------
    def reproject(self, spatial_reference, transformation=None, inplace=False):
        """
        Reprojects a given dataframe into a new coordinate system.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        spatial_reference               Required Integer/SpatialReference. The spatial reference the data
                                        should be reprojected into.
        ---------------------------     --------------------------------------------------------------------
        transformation                  Optional string. The optional transformation string.
        ---------------------------     --------------------------------------------------------------------
        inplace                         Optional boolean. Default False. Modify the SpatialDataFrame in
                                        place (do not create a new object)
        ===========================     ====================================================================

        :returns: SpatialDataFrame

        """
        if HASARCPY:
            if isinstance(spatial_reference, arcpy.SpatialReference):
                wkt = spatial_reference.exportToString()
                wkid = spatial_reference.factoryCode
                if wkid:
                    sr = _types.SpatialReference({'wkid' : wkid})
                elif wkt:
                    sr = _types.SpatialReference({'wkt': wkt})
                else:
                    sr = None
            elif isinstance(spatial_reference, int):
                sr = _types.SpatialReference({'wkid' : spatial_reference})
            elif isinstance(spatial_reference, string_types):
                sr = _types.SpatialReference({'wkt' : spatial_reference})
            elif isinstance(spatial_reference, _types.SpatialReference):
                sr = spatial_reference
            else:
                raise ValueError("spatial_referernce must be of type: int, string, _types.SpatialReference, or arcpy.SpatialReference")

            if inplace:
                df = self
            else:
                df = self.copy()
            sarcpy = sr.as_arcpy
            if sarcpy:
                geom = df.geometry.project_as(sarcpy, transformation)
                geom.sr = sr
                df.geometry = geom
                if inplace:
                    return df
            else:
                raise Exception("could not reproject the dataframe.")
            return df
    #----------------------------------------------------------------------
    def select_by_location(self, other, matches_only=True):
        """
        Selects all rows in a given SpatialDataFrame based on a given geometry

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        other                           Required Geometry. A geometry object to check for intersection.
        ---------------------------     --------------------------------------------------------------------
        matches_only                    Optional boolean.  if true, only matched records will be
                                        returned, else a field called 'select_by_location' will be added
                                        to the dataframe with the results of the select by location.
        ===========================     ====================================================================

        :returns: SpatialDataFrame

        """
        if isinstance(other, Geometry):
            if self.geometry_type.lower() == 'point':
                res = self.within(other)
            else:
                res = self.overlaps(other)
            if matches_only:
                return self[res]
            else:
                self['select_by_location'] = res
        else:
            raise ValueError("Input must be a geometry")
    #----------------------------------------------------------------------
    def merge_datasets(self, other):
        """
        This operation combines two dataframes into one new DataFrame.
        If the operation is combining two SpatialDataFrames, the
        geometry_type must match.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        other                           Required SpatialDataFrame. Another SpatialDataFrame to combine.
        ===========================     ====================================================================

        :returns: SpatialDataFrame

        """
        if isinstance(other, SpatialDataFrame) and \
           other.geometry_type == self.geometry_type:
            return pd.concat(objs=[self, other], axis=0)
        elif isinstance(other, DataFrame):
            return pd.concat(objs=[self, other], axis=0)
        elif isinstance(other, Series):
            self['merged_datasets'] = other
        elif isinstance(other, SpatialDataFrame) and \
             other.geometry_type != self.geometry_type:
            raise ValueError("Spatial DataFrames must have the same geometry type.")
        else:
            raise ValueError("Merge datasets cannot merge types %s" % type(other))
    #----------------------------------------------------------------------
    def erase(self, other, inplace=False):
        """
        Erases

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        other                           Required Geometry. A geometry object to erase from other geometries.
        ---------------------------     --------------------------------------------------------------------
        inplace                         Optional boolean. Default False. Modify the SpatialDataFrame in
                                        place (do not create a new object)
        ===========================     ====================================================================

        :returns: SpatialDataFrame
        """
        if inplace:
            df = self
        else:
            df = self.copy()
        if isinstance(other, Geometry):
            df.geometry = self.geometry.symmetricDifference(other)
            return df
        else:
            raise ValueError("Input must be of type arcpy.Geometry, not %s" % type(other))

###########################################################################
def _dataframe_set_geometry(self, col, drop=False, inplace=False, sr=None):
    if inplace:
        raise ValueError("Can't do inplace setting when converting from"
                         " DataFrame to SpatialDataFrame")
    gf = SpatialDataFrame(self)
    # this will copy so that BlockManager gets copied
    return gf.set_geometry(col, drop=drop, inplace=False, sr=sr)

if PY3:
    DataFrame.set_geometry = _dataframe_set_geometry
else:
    import types
    DataFrame.set_geometry = types.MethodType(_dataframe_set_geometry, None,
                                              DataFrame)



__version__ = '1.4.2'

from . import features, geoanalytics, geocoding, geometry, geoprocessing, network, raster, realtime, schematics, mapping, apps

from .gis import GIS
from .features.analysis import *
from .geocoding import geocode
#from .features._data.geodataset import SpatialDataFrame
__all__ = ['GIS', 'geocode', 'features',  'geoanalytics', 'geocoding', 'geometry', 'geoprocessing', 'network', 'raster', 'apps',
           'realtime', 'schematics', 'mapping',
             'aggregate_points',
             'calculate_density',
             'connect_origins_to_destinations',
             'create_buffers',
             'create_drive_time_areas',
             'create_route_layers',
             'create_viewshed',
             'create_watersheds',
             'derive_new_locations',
             'dissolve_boundaries',
             'enrich_layer',
             'extract_data',
             'find_existing_locations',
             'find_hot_spots',
             'find_nearest',
             'find_similar_locations',
             'interpolate_points',
             'join_features',
             'merge_layers',
             'overlay_layers',
             'plan_routes',
             'summarize_nearby',
             'summarize_within',
             'trace_downstream'
           ]

def _jupyter_nbextension_paths():
    return [{
        'section': 'notebook',
        'src': 'widgets',
        'dest': 'arcgis',
        'require': 'arcgis/mapview'
    }]




""" Defines the Assignment object.
"""
from .exceptions import ValidationError
from .feature_model import FeatureModel
from .managers import *
from ._schemas import AssignmentSchema


class Assignment(FeatureModel):
    """
    Represents an assignment

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    project                Required :class:`~arcgis.apps.workforce.Project`. The project that
                           this assignment belongs to.
    ------------------     --------------------------------------------------------------------
    feature                Optional :class:`~arcgis.features.Feature`.
                           A feature containing the assignments attributes. Mostly intended for
                           internal usage. If supplied, other parameters are ignored.
    ------------------     --------------------------------------------------------------------
    geometry               Optional :class:`Dict`.
                           A dictionary containing the assignment geometry
    ------------------     --------------------------------------------------------------------
    assignment_type        Optional :class:`~arcgis.apps.workforce.AssignmentType`.
                           The assignment type that represents this assignment.
    ------------------     --------------------------------------------------------------------
    assigned_date          Optional :class:`Date`
                           The date and time the assignment was assigned
    ------------------     --------------------------------------------------------------------
    assignment_read        Optional :class:`Bool`.
                           A flag indicating that the mobile worker has seen the assignment
    ------------------     --------------------------------------------------------------------
    completed_date         Optional :class:`Date`.
                           The date the assignment was completed
    ------------------     --------------------------------------------------------------------
    declined_comment       Optional :class:`String`.
                           The comment submitted by the mobile worker.
    ------------------     --------------------------------------------------------------------
    declined_date          Optional :class:`Date`.
                           The date the assignment was declined.
    ------------------     --------------------------------------------------------------------
    description            Optional :class:`Description`.
                           The description associated with the assignment.
    ------------------     --------------------------------------------------------------------
    dispatcher             Optional :class:`~arcgis.apps.workforce.Dispatcher`.
                           The dispatcher that assigned/created the assignment.
    ------------------     --------------------------------------------------------------------
    due_date               Optional :class:`Date`.
                           The date the assignment is due.
    ------------------     --------------------------------------------------------------------
    in_progress_date       Optional :class:`Date`.
                           The date the assignment was started.
    ------------------     --------------------------------------------------------------------
    location               Optional :class:`String`.
                           The location or address of the assignment.
    ------------------     --------------------------------------------------------------------
    notes                  Optional :class:`String`.
                           The notes associated with the assignment.
    ------------------     --------------------------------------------------------------------
    paused_date            Optional :class:`Date`.
                           The date and time the assignment was paused.
    ------------------     --------------------------------------------------------------------
    priority               Optional :class:`String`.
                           The priority of the assignment

                           `none`, `low`, `medium`, `high`, `critical`
    ------------------     --------------------------------------------------------------------
    status                 Optional :class:`String`.
                           The status of the assignment.

                           `unassigned`, `assigned`, `in_progress`, `completed`, `declined`,
                           `paused`, `canceled`
    ------------------     --------------------------------------------------------------------
    work_order_id          Optional :class:`String`.
                           The work order id associated with the assignment.
    ------------------     --------------------------------------------------------------------
    worker                 Optional :class:`~arcgis.apps.workforce.Worker`.
                           The worker assigned to the assignment
    ==================     ====================================================================
    """

    def __init__(self, project, feature=None, geometry=None, assignment_type=None,
            assigned_date=None, assignment_read=None, completed_date=None, declined_comment=None,
            declined_date=None, description=None, dispatcher=None, due_date=None, in_progress_date=None,
            location=None, notes=None, paused_date=None, priority=None, status=None,
            work_order_id=None, worker=None):
        super().__init__(project, project.assignments_layer, feature)
        self._schema = AssignmentSchema(project.assignments_layer)
        if feature:
            # create required objects from the feature
            # uses cached objects that are set when the assignments were last queried
            # speeds up construction by not querying FS for every assignment instantiated
            if feature.attributes[project._assignment_schema.worker_id]:
                self.worker = self.project._cached_workers[feature.attributes[project._assignment_schema.worker_id]]
            else:
                self.worker = None
            if feature.attributes[project._assignment_schema.dispatcher_id]:
                self.dispatcher = self.project._cached_dispatchers[feature.attributes[project._assignment_schema.dispatcher_id]]
            else:
                self.dispatcher = None
            if feature.attributes[project._assignment_schema.assignment_type]:
                self.assignment_type = self.project._cached_assignment_types[feature.attributes[project._assignment_schema.assignment_type]]
            else:
                self.assignment_type = None
        else:
            self.geometry = geometry
            self.assigned_date = assigned_date
            self.assignment_read = assignment_read
            self.completed_date = completed_date
            self.declined_comment = declined_comment
            self.declined_date = declined_date
            self.description = description
            self.due_date = due_date
            self.in_progress_date = in_progress_date
            self.location = location
            self.notes = notes
            self.paused_date = paused_date
            self.priority = priority
            self.status = status
            self.work_order_id = work_order_id
            self.assignment_type = assignment_type
            self.dispatcher = dispatcher
            self.worker = worker

    def __str__(self):
        type_name = self.assignment_type.name if self.assignment_type is not None else "no type"
        location = self.location if self.location is not None else "no location"
        return "{} at {}".format(type_name, location)

    def __repr__(self):
        return "<Assignment {}>".format(self.object_id)

    def update(self, geometry=None, assignment_type=None,
            assigned_date=None, assignment_read=None, completed_date=None, declined_comment=None,
            declined_date=None, description=None, dispatcher=None, due_date=None, in_progress_date=None,
            location=None, notes=None, paused_date=None, priority=None, status=None,
            work_order_id=None, worker=None):
        """
            Updates the assignment on the server

            ==================     ====================================================================
            **Argument**           **Description**
            ------------------     --------------------------------------------------------------------
            geometry               Optional :class:`Dict`.
                                   A dictionary containing the assignment geometry
            ------------------     --------------------------------------------------------------------
            assignment_type        Optional :class:`~arcgis.apps.workforce.AssignmentType`.
                                   The assignment type that represents this assignment.
            ------------------     --------------------------------------------------------------------
            assigned_date          Optional :class:`Date`
                                   The date and time the assignment was assigned
            ------------------     --------------------------------------------------------------------
            assignment_read        Optional :class:`Bool`.
                                   A flag indicating that the mobile worker has seen the assignment
            ------------------     --------------------------------------------------------------------
            completed_date         Optional :class:`Date`.
                                   The date the assignment was completed
            ------------------     --------------------------------------------------------------------
            declined_comment       Optional :class:`String`.
                                   The comment submitted by the mobile worker.
            ------------------     --------------------------------------------------------------------
            declined_date          Optional :class:`Date`.
                                   The date the assignment was declined.
            ------------------     --------------------------------------------------------------------
            description            Optional :class:`Description`.
                                   The description associated with the assignment.
            ------------------     --------------------------------------------------------------------
            dispatcher             Optional :class:`~arcgis.apps.workforce.Dispatcher`.
                                   The dispatcher that assigned/created the assignment.
            ------------------     --------------------------------------------------------------------
            due_date               Optional :class:`Date`.
                                   The date the assignment is due.
            ------------------     --------------------------------------------------------------------
            in_progress_date       Optional :class:`Date`.
                                   The date the assignment was started.
            ------------------     --------------------------------------------------------------------
            location               Optional :class:`String`.
                                   The location or address of the assignment.
            ------------------     --------------------------------------------------------------------
            notes                  Optional :class:`String`.
                                   The notes associated with the assignment.
            ------------------     --------------------------------------------------------------------
            paused_date            Optional :class:`Date`.
                                   The date and time the assignment was paused.
            ------------------     --------------------------------------------------------------------
            priority               Optional :class:`String`.
                                   The priority of the assignment

                                   `none`, `low`, `medium`, `high`, `critical`
            ------------------     --------------------------------------------------------------------
            status                 Optional :class:`String`.
                                   The status of the assignment.

                                   `unassigned`, `assigned`, `in_progress`, `completed`, `declined`,
                                   `paused`, `canceled`
            ------------------     --------------------------------------------------------------------
            work_order_id          Optional :class:`String`.
                                   The work order id associated with the assignment.
            ------------------     --------------------------------------------------------------------
            worker                 Optional :class:`~arcgis.apps.workforce.Worker`.
                                   The worker assigned to the assignment
            ==================     ====================================================================
        """
        update_assignment(self.project, self, geometry, assignment_type, assigned_date, assignment_read,
                          completed_date, declined_comment, declined_date, description,
                          dispatcher, due_date, in_progress_date, location, notes, paused_date,
                          priority, status, work_order_id, worker)

    def delete(self):
        """Deletes the assignment from the server"""
        delete_assignments(self.project, [self])

    @property
    def attachments(self):
        """Gets the :class:`~arcgis.apps.workforce.managers.AssignmentAttachmentManager` of the assignment"""
        return AssignmentAttachmentManager(self)

    @property
    def assigned_date(self):
        """Gets/Sets the assigned :class:`datetime` of the assignment"""
        return self._get_datetime_attr(self._schema.assigned_date)

    @assigned_date.setter
    def assigned_date(self, value):
        self._set_datetime_attr(self._schema.assigned_date, value)

    @property
    def assignment_read(self):
        """Gets/Sets the assignment read field"""
        return bool(self._feature.attributes.get(self._schema.assignment_read))

    @assignment_read.setter
    def assignment_read(self, value):
        self._feature.attributes[self._schema.assignment_read] = 1 if value else 0

    @property
    def assignment_type_code(self):
        """Gets the assignment type code"""
        return self._feature.attributes.get(self._schema.assignment_type)

    @property
    def assignment_type(self):
        """Gets/Sets the :class:`~arcgis.apps.workforce.AssignmentType`"""
        return self._assignment_type

    @assignment_type.setter
    def assignment_type(self, value):
        self._assignment_type = value
        if value:
            self._feature.attributes[self._schema.assignment_type] = value.code
        else:
            self._feature.attributes[self._schema.assignment_type] = None

    @property
    def completed_date(self):
        """Gets/Sets the completed :class:`datetime` of the assignment"""
        return self._get_datetime_attr(self._schema.completed_date)

    @completed_date.setter
    def completed_date(self, value):
        self._set_datetime_attr(self._schema.completed_date, value)

    @property
    def declined_comment(self):
        """Gets/Sets the declined comment of the assignment"""
        return self._feature.attributes.get(self._schema.declined_comment)

    @declined_comment.setter
    def declined_comment(self, value):
        self._feature.attributes[self._schema.declined_comment] = value

    @property
    def declined_date(self):
        """Gets/Sets the declined :class:`datetime` of the assignment"""
        return self._get_datetime_attr(self._schema.declined_date)

    @declined_date.setter
    def declined_date(self, value):
        self._set_datetime_attr(self._schema.declined_date, value)

    @property
    def description(self):
        """Gets/Sets the description for the assignment"""
        return self._feature.attributes.get(self._schema.description)

    @description.setter
    def description(self, value):
        self._feature.attributes[self._schema.description] = value

    @property
    def dispatcher_id(self):
        """Gets the dispatcher id of the assignment"""
        return self._feature.attributes.get(self._schema.dispatcher_id)

    @property
    def dispatcher(self):
        """Gets/Sets the :class:`~arcgis.apps.workforce.Dispatcher` of the assignment"""
        return self._dispatcher

    @dispatcher.setter
    def dispatcher(self, value):
        self._dispatcher = value
        self._feature.attributes[self._schema.dispatcher_id] = value.object_id if value is not None else None

    @property
    def due_date(self):
        """Gets/Sets the due :class:`datetime` of the assignment"""
        return self._get_datetime_attr(self._schema.due_date)

    @due_date.setter
    def due_date(self, value):
        self._set_datetime_attr(self._schema.due_date, value)

    @FeatureModel.geometry.setter
    def geometry(self, value):
        self._feature.geometry = value

    @property
    def in_progress_date(self):
        """Gets/Sets the in progress :class:`datetime` for the assignment"""
        return self._get_datetime_attr(self._schema.in_progress_date)

    @in_progress_date.setter
    def in_progress_date(self, value):
        self._set_datetime_attr(self._schema.in_progress_date, value)

    @property
    def location(self):
        """Gets/Sets the location of the assignment"""
        return self._feature.attributes.get(self._schema.location)

    @location.setter
    def location(self, value):
        self._feature.attributes[self._schema.location] = value

    @property
    def notes(self):
        """Gets/Sets the notes of the assignment"""
        return self._feature.attributes.get(self._schema.notes)

    @notes.setter
    def notes(self, value):
        self._feature.attributes[self._schema.notes] = value

    @property
    def paused_date(self):
        """Gets/Sets the paused :class:`datetime` for the assignment"""
        return self._get_datetime_attr(self._schema.paused_date)

    @paused_date.setter
    def paused_date(self, value):
        self._set_datetime_attr(self._schema.paused_date, value)

    @property
    def priority(self):
        """
            Gets/Sets the :class:`String` priority of the assignment

            `none`, `low`, `medium`, `high`, `critical`
        """
        lut = {
            0: "none",
            1: "low",
            2: "medium",
            3: "high",
            4: "critical"
        }
        if self._feature.attributes[self._schema.priority] is not None:
            return lut[self._feature.attributes[self._schema.priority]]
        else:
            return None

    @priority.setter
    def priority(self, value):
        if (isinstance(value, int) and value >=0 and value <= 4) or value is None:
            self._feature.attributes[self._schema.priority] = value
        elif isinstance(value, str):
            reduced_str = value.lower().replace(" ", "").replace("_", "")
            if reduced_str == "none":
                self._feature.attributes[self._schema.priority] = 0
            elif reduced_str == "low":
                self._feature.attributes[self._schema.priority] = 1
            elif reduced_str == "medium":
                self._feature.attributes[self._schema.priority] = 2
            elif reduced_str == "high":
                self._feature.attributes[self._schema.priority] = 3
            elif reduced_str == "critical":
                self._feature.attributes[self._schema.priority] = 4
            else:
                raise ValidationError("Invalid priority", self)
        else:
            raise ValidationError("Invalid priority", self)


    @property
    def status(self):
        """
            Gets/Sets the :class:`String` status of the assignment

           `unassigned`, `assigned`, `in_progress`, `completed`, `declined`,
           `paused`, `canceled`
        """
        lut = {
            0: "unassigned",
            1: "assigned",
            2: "in_progress",
            3: "completed",
            4: "declined",
            5: "paused",
            6: "canceled"
        }
        if self._feature.attributes[self._schema.status] is not None:
            return lut[self._feature.attributes[self._schema.status]]
        else:
            return None

    @status.setter
    def status(self, value):
        if (isinstance(value, int) and value >=0 and value <= 6) or value is None:
            self._feature.attributes[self._schema.status] = value
        elif isinstance(value, str):
            reduced_str = value.lower().replace(" ", "").replace("_", "")
            if reduced_str == "unassigned":
                self._feature.attributes[self._schema.status] = 0
            elif reduced_str == "assigned":
                self._feature.attributes[self._schema.status] = 1
            elif reduced_str == "inprogress":
                self._feature.attributes[self._schema.status] = 2
            elif reduced_str == "completed":
                self._feature.attributes[self._schema.status] = 3
            elif reduced_str == "declined":
                self._feature.attributes[self._schema.status] = 4
            elif reduced_str == "paused":
                self._feature.attributes[self._schema.status] = 5
            elif reduced_str == "canceled":
                self._feature.attributes[self._schema.status] = 6
            else:
                raise ValidationError("Invalid status", self)
        else:
            raise ValidationError("Invalid status", self)

    @property
    def work_order_id(self):
        """Gets/Sets the work order id of the assignment"""
        return self._feature.attributes.get(self._schema.work_order_id)

    @work_order_id.setter
    def work_order_id(self, value):
        self._feature.attributes[self._schema.work_order_id] = value

    @property
    def worker_id(self):
        """Gets the worker id of the assignment"""
        return self._feature.attributes.get(self._schema.worker_id)

    @property
    def worker(self):
        """Gets the :class:`~arcgis.apps.workforce.Worker` of the assignment"""
        return self._worker

    @worker.setter
    def worker(self, value):
        self._worker = value
        self._feature.attributes[self._schema.worker_id] = value.object_id if value is not None else None

    def _validate(self, **kwargs):
        errors = super()._validate(**kwargs)
        errors += self._validate_assignment_type()
        errors += self._validate_location()
        errors += self._validate_geometry()
        errors += self._validate_worker()
        errors += self._validate_dispatcher()
        errors += self._validate_status()
        return errors

    def _validate_for_add(self, **kwargs):
        errors = super()._validate_for_add(**kwargs)
        errors += self._validate_assignment_type_on_server()
        errors += self._validate_worker_on_server()
        errors += self._validate_dispatcher_on_server()
        return errors

    def _validate_for_update(self, **kwargs):
        errors = super()._validate_for_update(**kwargs)
        errors += self._validate_assignment_type_on_server()
        errors += self._validate_worker_on_server()
        errors += self._validate_dispatcher_on_server()
        return errors

    def _validate_assignment_type(self):
        errors = []
        if self.assignment_type is None:
            errors.append(ValidationError("An assignment must have an assignment type", self))
        elif self.assignment_type.code is None:
            errors.append(ValidationError("Invalid assignment type code: cannot be None", self))
        return errors

    def _validate_assignment_type_on_server(self):
        errors = []
        assignment_type = self.project._cached_assignment_types.get(self.assignment_type_code, None)
        if assignment_type is None:
            errors.append(ValidationError("Unrecognized assignment type", self))
        return errors

    def _validate_location(self):
        errors = []
        if not self.location or self.location.isspace():
            errors.append(ValidationError("Assignment cannot have an empty location", self))
        return errors

    def _validate_geometry(self):
        errors = []
        if not self.geometry:
            errors.append(ValidationError("Assignment must have geometry", self))
        return errors

    def _validate_worker(self):
        errors = []
        if self.worker is not None and self.worker.object_id is None:
            errors.append(ValidationError("Worker object_id cannot be None", self))
        return errors

    def _validate_worker_on_server(self):
        errors = []
        if self.worker is not None:
            worker = self.project._cached_workers.get(self.worker_id, None)
            if not worker:
                errors.append(ValidationError("Unrecognized worker object_id", self))
        return errors

    def _validate_dispatcher(self):
        errors = []
        if self.dispatcher is None:
            errors.append(ValidationError("An assignment must have a dispatcher", self))
        elif self.dispatcher.object_id is None:
            errors.append(ValidationError("Dispatcher object_id cannot be None", self))
        return errors

    def _validate_dispatcher_on_server(self):
        errors = []
        if self.dispatcher is not None:
            dispatcher = self.project._cached_dispatchers.get(self.dispatcher_id, None)
            if not dispatcher:
                errors.append(ValidationError("Unrecognized dispatcher object_id", self))
        return errors

    def _validate_status(self):
        errors = []
        if self.status is None:
            errors.append(ValidationError("Assignment status cannot be None", self))
        elif self.status == "unassigned":
            if self.worker is not None:
                message = "An UNASSIGNED assignment cannot have a worker"
                errors.append(ValidationError(message, self))
        elif self.status == "assigned":
            if self.worker is None:
                message = "An ASSIGNED assignment must have a worker"
                errors.append(ValidationError(message, self))
            if self.assigned_date is None:
                message = "An ASSIGNED assignment must have an assigned_date"
                errors.append(ValidationError(message, self))
        elif self.status == "in_progress":
            if self.worker is None:
                message = "An IN PROGRESS assignment must have a worker"
                errors.append(ValidationError(message, self))
            if self.assigned_date is None:
                message = "An IN PROGRESS assignment must have an assigned_date"
                errors.append(ValidationError(message, self))
            if self.in_progress_date is None:
                message = "An IN PROGRESS assignment must have an in_progress_date"
                errors.append(ValidationError(message, self))
        elif self.status == "paused":
            if self.worker is None:
                message = "A PAUSED assignment must have a worker"
                errors.append(ValidationError(message, self))
            if self.assigned_date is None:
                message = "A PAUSED assignment must have an assigned_date"
                errors.append(ValidationError(message, self))
            if self.in_progress_date is None:
                message = "A PAUSED assignment must have an in_progress_date"
                errors.append(ValidationError(message, self))
            if self.paused_date is None:
                message = "A PAUSED assignment must have a paused_date"
                errors.append(ValidationError(message, self))
        elif self.status == "completed":
            if self.worker is None:
                message = "A COMPLETED assignment must have a worker"
                errors.append(ValidationError(message, self))
            if self.assigned_date is None:
                message = "A COMPLETED assignment must have an assigned_date"
                errors.append(ValidationError(message, self))
            if self.in_progress_date is None:
                message = "A COMPLETED assignment must have an in_progress_date"
                errors.append(ValidationError(message, self))
            if self.completed_date is None:
                message = "A COMPLETED assignment must have a completed_date"
                errors.append(ValidationError(message, self))
        elif self.status == "declined":
            if self.worker is None:
                message = "A DECLINED assignment must have a worker"
                errors.append(ValidationError(message, self))
            if self.declined_date is None:
                message = "A DECLINED assignment must have a declined_date"
                errors.append(ValidationError(message, self))
            if self.declined_comment is None or self.declined_comment.isspace():
                message = "A DECLINED assignment must have a declined_comment"
                errors.append(ValidationError(message, self))
        return errors



"""
Represents messages left by users on a given Item in the GIS
"""
from __future__ import absolute_import
import json
from .connection import _ArcGISConnection
from ..gis import GIS, Item
########################################################################
class Comment(dict):
    """
    Represents messages left by users on a given Item in the GIS
    """
    _gis = None
    _portal = None
    _item = None
    _con = None
    _json_dict = None
    _json = None
    _hydrated = None
    def __init__(self, url, item, data=None, initialize=True, **kwargs):
        """
        class initializer

        Parameters:
         :url: web address to the resource
         :item: arcgis.gis.Item class where the comment originate from
         :data: allows the object to be pre-populated with information from
          a dictionary
         :initialize: if True, on creation, the object will hydrate itself.
        """
        super(Comment, self).__init__()
        self._url = url
        self._gis = item._gis
        self._portal = self._gis._portal
        if isinstance(item._gis, _ArcGISConnection):
            self._con = item._gis
        elif isinstance(item._gis, GIS):
            self._gis = item._gis
            self._con = item._gis._con
        else:
            raise ValueError(
                "connection must be of type GIS or _ArcGISConnection")
        if data and \
           isinstance(data, dict):
            for k,v in data.items():
                self[k] = v
            self.__dict__.update(data)
        if initialize:
            self._init(connection=self._con)
            self._hydrated = True
        else:
            self._hydrated = False
    #----------------------------------------------------------------------
    def _init(self, connection=None):
        """loads the properties into the class"""
        if connection is None:
            connection = self._con
        attributes = [attr for attr in dir(self)
                      if not attr.startswith('__') and \
                      not attr.startswith('_')]
        params = {"f":"json"}
        result = connection.get(path=self._url,
                                params=params)
        self._json_dict = result
        for k,v in result.items():
            if k in attributes:
                setattr(self, "_"+ k, v)
                self[k] = v
            else:
                self[k] = v
        self.__dict__.update(result)
    #----------------------------------------------------------------------
    def __getattr__(self, name):
        if not self._hydrated and not name.startswith('_'):
            self._init()
        try:
            return self.__dict__[name]
        except:
            raise AttributeError("'%s' object has no attribute '%s'" % (type(self).__name__, name))
    #----------------------------------------------------------------------
    def __getitem__(self, k):
        try:
            return self.__dict__[k]
        except KeyError:
            if not self._hydrated and not k.startswith('_'):
                self._init()
            return self.__dict__[k]
    #----------------------------------------------------------------------
    def update(self, comment):
        """
        Updates a comment. Available only to the authenticated user who
        created the comment.

        Parameters:
         :comment: updated comment text
        Output:
         On successful update, the comment Id
         On unsuccessful update, JSON response message
        """
        url = "%s/update" % self._url
        params = {"f" : "json",
                  "comment" : comment}
        res = self._con.post(path=url,
                             postdata=params)
        if 'commentId' in res:
            self._init()
            return res['commentId']
        return res
    #----------------------------------------------------------------------
    def delete(self):
        """
        removes a given comment
        """
        url = "%s/delete" % self._url
        params = {"f" : "json"}
        res = self._con.post(path=url,
                             postdata=params)
        if 'success' in res:
            return res['success']
        return res

